#interface objects
from .InterfaceObjects import Parameter,PerformanceModel,StopCondition,Constraint,Objective,DecisionVariable
#import helper functions
from .HelperFunctions import *

#-----------------------------------------------------------------------------------------------------------
#
#   Performance extension
#
#
# -----------------------------------------------------------------------------------------------------------
def getPerformance( OptimizationProblemName,model):
    """
      Function to fetch the complete performance model using the name as identifier.
      The function returns an interface block of the performance model

      :param string OptimizationProblemName: Name identifier of the optimization problem
      :param object model: Metamodel instance model

      :return object PerformanceModel: Interface object of the performance model (-1 = error)
    """

    #define performance model interface object
    pModel = PerformanceModel()
    pModel.clean()  # Flushing the pModel

    found = 0
    if len(model.includesOptimizationProblem.items) != 0:
        # match the assemblySystem by Name
        for optimizationProblem in model.includesOptimizationProblem.items:
            if OptimizationProblemName == optimizationProblem.hasName:
                found = 1
                pModel.Name=optimizationProblem.hasName
                pModel.Description = optimizationProblem.hasDescription

                # fetch the analysysDescription
                pModel.OptimizationMethod=[optimizationProblem.hasOptimizationAnalysisdDescription.hasOptimizationMethod.hasName,optimizationProblem.hasOptimizationAnalysisdDescription.hasOptimizationMethod.hasDescription,optimizationProblem.hasOptimizationAnalysisdDescription.hasOptimizationMethod.hasAlgorithmClass.name]
                StopList = []
                for STP in optimizationProblem.hasOptimizationAnalysisdDescription.hasStopCriterion.items:
                    s = StopCondition(Name=STP.hasName,Description=STP.hasDescription,Value=STP.hasValue,Stopcriteria=STP.hasStopCriterion.name)
                    StopList.append(s)
                pModel.StopConditionList=StopList

                # fetch the Optimization targets
                for designTarget in optimizationProblem.hasOptimizationTargetDescription.hasDesignTarget.items:
                    # check the typing
                    x = str(type(designTarget))
                    if 'Constraint' in x:
                        cstr = Constraint(Name=designTarget.hasName,Description=designTarget.hasDescription)

                        for term in designTarget.hasExpression.hasTerm.items:                       #TODO:this needs to be reworked!
                            TERM = []
                            if "LEFT" in term.hasInfixPosition.name:
                                TERM.append('LEFT')
                            if "RIGHT" in term.hasInfixPosition.name:
                                TERM.append('RIGHT')
                            x = str(type(term))
                            if 'Textual' in x:  # textual term
                                TERM.append(term.hasValue)
                            if 'Numerical' in x:  # numerical term
                                TERM.append(term.hasValue)
                            if 'Variable' in x:  # variable term
                                TERM.append(term.hasDecisionVariable.hasName)
                                TERM.append(term.hasDecisionVariable.hasDescription)
                                TERM.append(term.hasDecisionVariable.hasOptimum)
                            cstr.Expression.append(TERM)
                        #add the expression operator
                        operator  = designTarget.hasExpression.hasExpressionOperator.hasValue
                        cstr.Expression.append(operator)

                        pModel.ConstraintList.append(cstr)

                    if 'Objective' in x:
                        OBJ_O = Objective(Name=designTarget.hasName, Description=designTarget.hasDescription)

                        x = str(type(designTarget.hasTerm))
                        if 'Textual' in x:  # textual term
                            OBJ_O.ObjectiveOption = designTarget.hasTerm.hasValue
                        if 'Numerical' in x:  # numerical term
                            OBJ_O.ObjectiveOption = designTarget.hasTerm.hasValue
                        if 'Variable' in x:  # variable term
                            OBJ_O.ObjectiveOption = [designTarget.hasTerm.hasDecisionVariable.hasName,designTarget.hasTerm.hasDecisionVariable.hasDescription,designTarget.hasTerm.hasDecisionVariable.hasOptimum]
                        pModel.ObjectiveList.append(OBJ_O)

                # fetch the design variables
                for variable in optimizationProblem.hasOptimizationTargetDescription.hasDecisionVariable.items:
                    dv = DecisionVariable(Name=variable.hasName,Description=variable.hasDescription,InitialValue=variable.hasInitialValue,MaxValue=variable.hasMaxValue,MinValue=variable.hasMinValue,Optimum=variable.hasOptimum,Resolution=variable.hasResolution)
                    try:
                        par = Parameter(Name=variable.hasParameter.hasName,Description=variable.hasParameter.hasDescription,Key=variable.hasParameter.hasKey, GUID=variable.hasParameter.hasGUID,Value=variable.hasParameter.hasValue)
                        dv.parameterList.append(par)
                    except:
                        par = []

                    pModel.parameterList.append(dv)



        if found:
            return pModel
        else:
            return -1
    else:
        return -1  # no Optimization Problem!


def updatePerformance(interfaceObject,model,KBPath):
    """
          Function to update the complete performance model matching the name as identifier.
          The performance interface object is used to interface with the function.
          The function returns whether or not the function is performed correctly

          :param object PerformanceModel: Interface object of the performance model
          :param object model: Metamodel instance model
          :param string KBPath: Absolute path to the metamodel instance model

          :return int Error: -1 = error, 1= function performed correcly
        """

    #find the optimization model
    found = 0
    if len(model.includesOptimizationProblem.items) != 0:
        # match the assemblySystem by Name
        for optimizationProblem in model.includesOptimizationProblem.items:
            if interfaceObject.Name == optimizationProblem.hasName:
                found = 1
                optimizationProblem.hasName=interfaceObject.Name
                optimizationProblem.hasDescription=interfaceObject.Description
                #analysisDescription
                optimizationProblem.hasOptimizationAnalysisdDescription.hasOptimizationMethod.hasName = interfaceObject.OptimizationMethod[0]
                optimizationProblem.hasOptimizationAnalysisdDescription.hasOptimizationMethod.hasDescription = interfaceObject.OptimizationMethod[1]
                optimizationProblem.hasOptimizationAnalysisdDescription.hasOptimizationMethod.hasAlgorithmClass.name = interfaceObject.OptimizationMethod[2]

                StopList = []
                for STP in optimizationProblem.hasOptimizationAnalysisdDescription.hasStopCriterion.items:
                    for STP_NEW in interfaceObject.StopConditionList:
                        if STP_NEW.Name == STP.hasName:
                            STP.hasName=STP_NEW.Name
                            STP.hasDescription=STP_NEW.Description
                            STP.hasValue=STP_NEW.Value
                            STP.hasStopCriterion.name=STP_NEW.StopCriteria

                #Objectives and constraints
                for designTarget in optimizationProblem.hasOptimizationTargetDescription.hasDesignTarget.items:
                    # check the typing
                    x = str(type(designTarget))
                    if 'Constraint' in x:
                        for constraint_new in interfaceObject.ConstraintList:
                            if constraint_new.Name ==designTarget.hasName:
                                designTarget.hasName=constraint_new.Name
                                designTarget.hasDescription=constraint_new.Description
                                for term in designTarget.hasExpression.hasTerm.items:
                                    y=1
                                    #TODO: using the current KB (version6), updating the expressions is not possible! discuss if needed


                    if 'Objective' in x:
                        for obj_new in interfaceObject.ObjectiveList:
                            designTarget.hasName = obj_new.Name
                            designTarget.hasDescription=obj_new.Description
                        x = str(type(designTarget.hasTerm))
                        if 'Textual' in x:  # textual term
                            designTarget.hasTerm.hasValue = obj_new.ObjectiveOption[0]
                        if 'Numerical' in x:  # numerical term
                            designTarget.hasTerm.hasValue = obj_new.ObjectiveOption[0]
                        if 'Variable' in x:  # variable term
                            designTarget.hasTerm.hasDecisionVariable.hasName= obj_new.ObjectiveOption[0]
                            designTarget.hasTerm.hasDecisionVariable.hasDescription= obj_new.ObjectiveOption[1]
                            designTarget.hasTerm.hasDecisionVariable.hasOptimum = obj_new.ObjectiveOption[2]



                #Decision Variables
                for variable in optimizationProblem.hasOptimizationTargetDescription.hasDecisionVariable.items:
                    for variable_new in interfaceObject.parameterList:
                        if variable.hasName == variable_new.Name:
                            variable.hasName=variable_new.Name
                            variable.hasInitialValue=variable_new.InitialValue
                            variable.hasMaxValue=variable_new.MaxValue
                            variable.hasMinValue=variable_new.MinValue
                            variable.hasOptimum=variable_new.Optimum
                            variable.hasResolution=variable_new.Resolution
                            for par in variable_new.parameterList:
                                if variable.hasParameter.hasName==par.Name:
                                    variable.hasParameter.hasName = par.Name
                                    variable.hasParameter.hasDescription = par.Description
                                    variable.hasParameter.hasKey = par.Key
                                    variable.hasParameter.hasGUID = par.GUID
                                    variable.hasParameter.hasValue = str(par.Value)


        if found:
            updateKBv6(KBPath, model)
            return 1
        else:
            return -1
    else:
        return -1  # no Optimization Problem!


def setPerformance(interfaceObject,model,KBPath,API):
    """
             Function to add a complete performance model to an blank KB (no existing performance part).
             The performance interface object is used to interface with the function.
             The function returns whether or not the function is performed correctly

             :param object PerformanceModel: Interface object of the performance model
             :param object model: Metamodel instance model
             :param string KBPath: Absolute path to the metamodel instance model

             :return int Error: -1 = error, 1= function performed correcly
    """
    # generate a new ASG config and configure
    optimizationProblem = API.create_connected("OptimizationProblem")
    optimizationProblem.hasName=interfaceObject.Name
    optimizationProblem.hasDescription=interfaceObject.Description
    #analysisDescription
    OptimizationAnalysisdDescription = API.create_noPlatformRoot("OptimizationAnalysisDescription")

    #add the method
    OptimizationMethod = API.create_noPlatformRoot("OptimizationMethod")
    OptimizationMethod.hasName = interfaceObject.OptimizationMethod[0]
    OptimizationMethod.hasDescription = interfaceObject.OptimizationMethod[1]
    OptimizationMethod.hasAlgorithmClass.name = interfaceObject.OptimizationMethod[2]
    OptimizationAnalysisdDescription.hasOptimizationMethod = OptimizationMethod

    #add the stop criteria
    for STP_NEW in interfaceObject.StopConditionList:
        StopCriterion = API.create_noPlatformRoot("StopCriterion")
        StopCriterion.hasName = STP_NEW.Name
        StopCriterion.hasDescription = STP_NEW.Description
        StopCriterion.hasValue = STP_NEW.Value
        StopCriterion.hasStopCriterion.name = STP_NEW.StopCriteria
        OptimizationAnalysisdDescription.hasStopCriterion.append(StopCriterion)

    optimizationProblem.hasOptimizationAnalysisdDescription = OptimizationAnalysisdDescription

    OptimizationTargetDescription = API.create_noPlatformRoot("OptimizationTargetDescription")
    #TODO:implement!
    #Objectives and constraints
    for constraint_new in interfaceObject.ConstraintList:
        constraint = API.create_noPlatformRoot("Constraint")
        constraint.hasName=constraint_new.Name
        constraint.hasDescription=constraint_new.Description

        OptimizationTargetDescription.hasDesignTarget.append(constraint)

    for obj_new in interfaceObject.ObjectiveList:
        objective = API.create_noPlatformRoot("Objective")
        objective.hasName = obj_new.Name
        objective.hasDescription=obj_new.Description
        expression = API.create_noPlatformRoot("Expression")

        # TODO: this needs to be updated via the object - no example available to validate
        #expression.hasExpressionOperator = "<"
        #expression.hasValue = double(obj_new.ObjectiveOption[2])
        #expression.hasTerm.append(obj_new.ObjectiveOption[1])
        #objective.hasExpression #TODO: using the current KB (version6), updating the expressions (hasExpression) is not possible! discuss if needed

        OptimizationTargetDescription.hasDesignTarget.append(objective)


    #Decision Variables
    for variable_new in interfaceObject.parameterList:
        variable = API.create_noPlatformRoot("Variable")
        variable.hasName=variable_new.Name
        variable.hasInitialValue=variable_new.InitialValue
        variable.hasMaxValue=variable_new.MaxValue
        variable.hasMinValue=variable_new.MinValue
        variable.hasOptimum=variable_new.Optimum
        variable.hasResolution=variable_new.Resolution
        for par in variable_new.parameterList:
            subvariable = API.create_noPlatformRoot("Variable")
            subvariable.hasName = par.Name
            subvariable.hasDescription = par.Description
            subvariable.hasKey = par.Key
            subvariable.hasGUID = par.GUID
            subvariable.hasValue = str(par.Value)

        OptimizationTargetDescription.hasDecisionVariable.append(variable)

    optimizationProblem.hasOptimizationTargetDescription = OptimizationTargetDescription

    updateKBv6(KBPath, model)
    return 1



###################################################################################################
#
#   @brief: this class contains KB interface functions (API)
#
#
#   @Context: developed for the PACo project
#
#
#   @Brief : Basic API function library
#                       -
#
###################################################################################################
#          @ Author: Bert Van Acker
#
#                              revision
#   date             Who rel ver              Remarks
#  28/06/2021        BVA v0.1                 Setup basic structure
#  13/12/2021        BVA v0.2                 KB v6 support
#  21/02/2022        BVA v0.3                 create empty KB function added
###################################################################################################
#imports
import os
import re
from pyecore.ecore import EClass
from pyecore.resources import ResourceSet, URI
from pyecore.resources.xmi import XMIResource
from pyecore.utils import DynamicEPackage

#import helper functions
from .HelperFunctions import *
#import ASG functions
from .ASGFunctions import *
#import DFA functions
from .DFAFunctions import *
#import DFA functions
from .PerformanceFunctions import *
#import Contract functions
from .ContractFunctions import *

#interface objects
from .InterfaceObjects import ASG,StopCondition,DFARule,Product,Parameter,ProductPart,AssemblySequence,Operation,Operator,Tool,AssemblyOption



class KB_Interface():

    def __init__(self,DEBUG):
        self.DEBUG = DEBUG
        self.MM = None
        self.model_instance = None
        self.model = None
        self.KB_path = ""
        self.ECORE_path = ""

# -----------------------------------------------------------------------------------------------------------
#
#   General functions
#
# -----------------------------------------------------------------------------------------------------------
    def createEmptyKB(self,Name="PACo_KB",OutputPath="output/"):
        """
                Function to create an empty Knowledge Base

                :param string Name: Name identifier of the assemlbySystem
                :param string OutputPath: path to the output folder

                :return string KB instance path: Path to the created KB instance (-1 = error)
          """
        KB_path = OutputPath+Name+".pacopackage"
        f = open(KB_path, "w+")
        f.write("<?xml version='1.0' encoding='UTF-8'?>\n")
        f.write('<paco:PlatformRoot xmlns:xmi="http://www.omg.org/XMI" xmlns:paco="http://www.ua.be/cosys/paco/PACoMM" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmi:version="2.0">\n')
        f.write('  <includesOperator/>\n')  #HACK: this is a placeholder for future empty class problems
        f.write("</paco:PlatformRoot>\n")
        return KB_path


#-----------------------------------------------------------------------------------------------------------
#
#   ASG extension
#
# -----------------------------------------------------------------------------------------------------------
    def getASG(self,AssemblySystemName):
        """
              Function to fetch the Assembly Sequence Generation (ASG) configuration using the name as identifier.
              The function returns an interface block of the ASG

              :param string AssemblySystemName: Name identifier of the assemlbySystem

              :return object ASGModel: Interface object of the ASG (-1 = error)
        """
        return getASG(AssemblySystemName,self.model,self.KB_path)

    def updateASG(self,AssemblySystemName, interfaceObject):
        """
              Function to update the  complete Assembly Sequence Generation (ASG) configuration using the name as identifier.
              The ASG interface object is used to interface with the function.
              The function returns whether or not the function is performed correctly

              :param string AssemblySystemName: Name identifier of the assemlbySystem
              :param object ASGModel: Interface object of the ASG

              :return int Error: -1 = error, 1= function performed correcly
        """

        return updateASG(AssemblySystemName,interfaceObject, self.model, self.KB_path)

    def setASG(self, AssemblySystemName, InterfaceObject):
        """
              Function to add a complete ASG model to an blank KB (no existing ASG part).
              The function returns whether or not the function is performed correctly (NOT YET)

              :param string AssemblySystemName: Name identifier of the assemlbySystem
              :param object ASGModel: Interface object of the ASGModel model

              :return int Error: -1 = error, 1= function performed correcly
        """
        setASG(AssemblySystemName, InterfaceObject,self.model,self.KB_path,self.ECORE_path,self.MM,self)

# -----------------------------------------------------------------------------------------------------------
#
#   DFA extension
#
# -----------------------------------------------------------------------------------------------------------
    def getASG_DFARules(self, AssemblySystemName, component):
        """
             Function to fetch the ASG contained rules using the name as identifier and component as selection mechanism.
             The function returns object list of DFA rules

             :param string AssemblySystemName: Name identifier of the assemlbySystem
             :param string component: Identify the rule container (Selector/Evaluator)

             :return objectlist DFARules: List of DFA rules (-1 = error)
        """
        return getASG_DFARules(AssemblySystemName, component,self.model)

    def updateASG_DFARules(self, AssemblySystemName, component,rule):
        """
             Function to update a single the DFA rules within the given container.
             The function returns whether or not the function is performed correctly (NOT YET)

             :param string AssemblySystemName: Name identifier of the assemlbySystem
             :param object ruleContainer: container object containing the DFA rules
             :param object rule: The DFA rules

             :return int Error: -1 = error, 1= function performed correcly
        """
        return updateASG_DFARules(AssemblySystemName, component,rule,self.model,self.KB_path)

    def getDFArules(self,ruleContainer):
        """
             Function to fetch the DFA rules within the given container.
             The function returns object list of DFA rules

             :param object ruleContainer: container object containing the DFA rules

             :return objectlist DFARules: List of DFA rules (-1 = error)
        """
        return getDFArules(ruleContainer)

    def updateDFArule(self, ruleContainer,rule):
        """
             Function to update a single the DFA rules within the given container.
             The function returns whether or not the function is performed correctly (NOT YET)

             :param object ruleContainer: container object containing the DFA rules
             :param object rule: The DFA rules

             :return int Error: -1 = error, 1= function performed correcly
        """
        updateDFArule( ruleContainer,rule)

    def addASG_DFARule(self, AssemblySystemName, component, rule):
        """
             Function to update a single the DFA rules within the given container.
             The function returns whether or not the function is performed correctly (NOT YET)

             :param string AssemblySystemName: Name identifier of the assemlbySystem
             :param object ruleContainer: container object containing the DFA rules
             :param object rule: The DFA rules

             :return int Error: -1 = error, 1= function performed correcly
        """
        addASG_DFARule(AssemblySystemName,component,rule,self.model,self.KB_path,self)

    def addASG_DFARules(self, AssemblySystemName, component, ruleList):
        """
             Function to update multiple DFA rules within the given container.
             The function returns whether or not the function is performed correctly (NOT YET)

             :param string AssemblySystemName: Name identifier of the assemlbySystem
             :param object ruleContainer: container object containing the DFA rules
             :param object rule: The DFA rules

             :return int Error: -1 = error, 1= function performed correcly
        """
        addASG_DFARules(AssemblySystemName,component,ruleList,self.model,self.KB_path,self)
# -----------------------------------------------------------------------------------------------------------
#
#   PERFORMANCE extension
#
# -----------------------------------------------------------------------------------------------------------

    def getPerformance(self, OptimizationProblemName):
        """
          Function to fetch the complete performance model using the name as identifier.
          The function returns an interface block of the performance model

          :param string OptimizationProblemName: Name identifier of the optimization problem

          :return object PerformanceModel: Interface object of the performance model (-1 = error)
        """
        return getPerformance(OptimizationProblemName,self.model)

    def updatePerformance(self, interfaceObject):
        """
              Function to update the complete performance model matching the name as identifier.
              The performance interface object is used to interface with the function.
              The function returns whether or not the function is performed correctly

              :param object PerformanceModel: Interface object of the performance model

              :return int Error: -1 = error, 1= function performed correcly
        """
        return updatePerformance(interfaceObject,self.model,self.KB_path)

    def setPerformance(self, interfaceObject):
        """
             Function to add a complete performance model to an blank KB (no existing performance part).
             The performance interface object is used to interface with the function.
             The function returns whether or not the function is performed correctly

             :param object PerformanceModel: Interface object of the performance model

             :return int Error: -1 = error, 1= function performed correcly
        """
        return setPerformance(interfaceObject,self.model,self.KB_path,self)



# -----------------------------------------------------------------------------------------------------------
#
#   Product (request Pavel)
#           TODO: to add parameters, we need to specify the model, can we add this to the API signature?
# -----------------------------------------------------------------------------------------------------------
    def getProduct(self):
        """
         Function to fetch the Product .
         The function returns an interface block of the Product


         :return object ProductModel: Interface object of the Product model (-1 = error)
        """

        if len(self.model.includesProduct.items) != 0:
            product = Product(Name=self.model.includesProduct.items[0].hasProductName,STEPFile=self.model.includesProduct.items[0].hasSTEPfile)
            #TODO: add the parameters-> discuss which parameters
            return product

        else:
            return -1

    def updateProduct(self,interfaceObject):
        """
              Function to update the Product matching the name as identifier.
              The Product interface object is used to interface with the function.
              The function returns whether or not the function is performed correctly

              :param object ProductModel: Interface object of the Product model

              :return int Error: -1 = error, 1= function performed correcly
        """

        if len(self.model.includesProduct.items) != 0:
            if interfaceObject.Name ==self.model.includesProduct.items[0].hasProductName:
                self.model.includesProduct.items[0].hasProductName=interfaceObject.Name
                self.model.includesProduct.items[0].hasSTEPfile=interfaceObject.STEPFile

                self.updateKBv6()
                return 1
            else:
                return -1

        else:
            return -1

    def setProduct(self, interfaceObject):
        """
             Function to add a Product model to an blank KB .
             The Product interface object is used to interface with the function.
             The function returns whether or not the function is performed correctly

             :param object ProductModel: Interface object of the Product model

             :return int Error: -1 = error, 1= function performed correcly
        """
        if len(self.model.includesProduct.items) != 0:
            #Product already exists!
            print("Product already exists, use updateProduct() instead")
            return -1
        else:
            # --!HACK!--:check if an Optimization class already exists, if not, generate a placeholder
            #self.model = EmptyClassHack(API=self, elementName="Product")

            product = self.create_connected("Product")
            product.hasProductName = interfaceObject.Name
            product.hasSTEPfile = interfaceObject.STEPFile
            # TODO: discuss what is set via the setProduct function e.g. do we already set the analysis model?

            self.updateKBv6()
            return 1

# -----------------------------------------------------------------------------------------------------------
#
#   Product part
#
# -----------------------------------------------------------------------------------------------------------
    def getProductPart(self,ProductPartName):
        """
         Function to fetch the ProductPart using the name as identifier .
         The function returns an interface block of the ProductPart

        :param string ProductPartName: Name identifier of the ProductPart

         :return object ProductPartModel: Interface object of the ProductPart model (-1 = error)
        """
        if len(self.model.includesProductPart.items) != 0:
            for productPart in self.model.includesProductPart.items:
                if ProductPartName == productPart.hasPartName:
                    part = ProductPart(Name=productPart.hasPartName,ProductType=productPart.hasPartType,Quantity=1)        #TODO:Quantity not in MM v6
                    #link the parameters
                    part.parameterList = []
                    try:
                        for par in productPart.hasModel.hasModelParameter.items:
                            p = Parameter(Name='Not defined',Key=par.hasKey,Value=par.hasValue,Stopcriteria='Not defined')  #TODO: no complete parameter description is given!
                            part.parameterList.append(p)
                    except:
                        if self.DEBUG:print("No model added to ProductPart")

                    #link the product parts - Recursive!
                    part.partList = []
                    for subPart in productPart.hasCompositionEdge.items:
                        p = self.getProductPart(subPart.hasPartName)
                        part.partList.append(p)

                    #link the contactFeatures
                    #TODO: not yet implemented - No example available to validate

                    #material
                    try:
                        part.material = [productPart.hasMaterial.hasName,productPart.hasMaterial.hasCost,productPart.hasMaterial.hasDamping,productPart.hasMaterial.hasDensity,productPart.hasMaterial.hasTransparency,productPart.hasMaterial.hasStiffness.hasScalarMatrix6x6]
                    except:
                        if self.DEBUG: print("No material added to ProductPart")

                    return part

        else:
            return -1  # no ProductPart instance

    def updateProductPart(self,interfaceObject):
        """
              Function to update the ProductPart matching the name as identifier.
              The ProductPart interface object is used to interface with the function.
              The function returns whether or not the function is performed correctly

              :param object ProductPartModel: Interface object of the ProductPart model

              :return int Error: -1 = error, 1= function performed correcly
        """
        if len(self.model.includesProductPart.items) != 0:
            for productPart in self.model.includesProductPart.items:
                if interfaceObject.Name == productPart.hasPartName:
                    productPart.hasPartName=interfaceObject.Name
                    #productPart.hasDescription = interfaceObject.Description       # TODO:Quantity not in MM v6
                    productPart.hasPartType=interfaceObject.ProductType
                    #productPart.Quantity = interfaceObject.Quantity                 # TODO:Quantity not in MM v6

                    # update the parameters
                    for par in productPart.hasModel.hasModelParameter.items:
                        for par_new in interfaceObject.parameterList:
                            if par_new.Key == par.hasKey:
                                par.hasKey = par_new.Key
                                par.hasValue = par_new.Value


                    # link the product parts - Recursive!
                    for par in productPart.hasCompositionEdge.items:
                        for par_new in interfaceObject.partList:
                            if par_new.Name == par.hasPartName:
                                #TODO:update the subproduct
                                error = self.updateProductPart(par_new)


                    # link the contactFeatures
                    #TODO: not yet implemented - No example available to validate

                    # material
                    productPart.hasMaterial.hasName=interfaceObject.material[0]
                    productPart.hasMaterial.hasCost=interfaceObject.material[1]
                    productPart.hasMaterial.hasDamping=interfaceObject.material[2]
                    productPart.hasMaterial.hasDensity=interfaceObject.material[3]
                    productPart.hasMaterial.hasTransparency=interfaceObject.material[4]
                    productPart.hasMaterial.hasStiffness.hasScalarMatrix6x6=interfaceObject.material[5]

                    self.updateKBv6()
                    return 1

        else:
            return -1  # no ProductPart instance

    def setProductPart(self, interfaceObject):
        """
             Function to add a ProductPart model to an blank KB .
             The ProductPart interface object is used to interface with the function.
             The function returns whether or not the function is performed correctly

             :param object ProductPartModel: Interface object of the ProductPart model

             :return int Error: -1 = error, 1= function performed correcly
        """

        #setup a eCore class of a productPart
        part = self.create_connected("ProductPart")

        part.hasPartName = interfaceObject.Name
        # productPart.hasDescription = interfaceObject.Description       # TODO:Quantity not in MM v6
        part.hasPartType = interfaceObject.ProductType
        # productPart.Quantity = interfaceObject.Quantity                 # TODO:Quantity not in MM v6


        #add model to ProductPart
        model = self.create_connected("Model")
        model.hasModelType = ""
        part.hasModel = model

        # update the parameters
        for par_new in interfaceObject.parameterList:
            parameter = self.create_connected("Parameter")
            parameter.hasKey = par_new.Key
            parameter.hasValue = par_new.Value
            part.hasModel.hasModelParameter.append(parameter)

        # link the product parts - Recursive! # TODO:update the subproduct - No example available to validate
        for par_new in interfaceObject.partList:
            if par_new.Name == part.hasPartName:
                error = self.setProductPart(par_new)
                part.hasCompositionEdges.append(self.model.includesProductPart.items[-1])

        # link the contactFeatures
        # TODO: not yet implemented - No example available to validate

        # material
        material = self.create_connected("Material")
        part.hasMaterial = material
        part.hasMaterial.hasName = interfaceObject.material[0]
        part.hasMaterial.hasCost = interfaceObject.material[1]
        part.hasMaterial.hasDamping = interfaceObject.material[2]
        part.hasMaterial.hasDensity = interfaceObject.material[3]
        part.hasMaterial.hasTransparency = interfaceObject.material[4]
        stiffness = self.create_connected("Stiffness")
        part.hasMaterial.hasStiffness = stiffness
        part.hasMaterial.hasStiffness.hasScalarMatrix6x6 = interfaceObject.material[5]

        self.model.includesProduct[-1].hasProductPart.append(part)
        
        self.updateKBv6()

        return 1


    def getAllProductParts(self):
        """
         Function to fetch all ProductParts within the KB .
         The function returns a list of interface block of the ProductPart


         :return objectlist ProductPartList: List of Interface object of the ProductPart model (-1 = error)
        """
        if len(self.model.includesProductPart.items) != 0:
            list = []
            for productPart in self.model.includesProductPart.items:
                p=self.getProductPart(productPart.hasPartName)
                list.append(p)

            return list
        else:
            return -1

    def updateAllProductParts(self, interfaceObjectList):
        """
              Function to update all ProductParts within the KB.
              A list of ProductPart interface objects is used to interface with the function.
              The function returns whether or not the function is performed correctly

              :param objectlist interfaceObjectList: List of interface object of the ProductPart model

              :return int Error: -1 = error, 1= function performed correcly
        """
        error_overall = 1
        for interfaceObject in interfaceObjectList:
            error = self.updateProductPart(interfaceObject)
            if error==-1:
                error_overall = -1

        return error_overall

    def setAllProductParts(self, interfaceObjectList):
        """
             Function to add all ProductParts to an blank KB .
             A list of ProductPart interface objects is used to interface with the function.
             The function returns whether or not the function is performed correctly

             :param objectlist interfaceObjectList: List of interface object of the ProductPart model

             :return int Error: -1 = error, 1= function performed correcly
        """
        error_overall = 1
        for interfaceObject in interfaceObjectList:
            error = self.setProductPart(interfaceObject)
            if error == -1:
                error_overall = -1

        return error_overall

# -----------------------------------------------------------------------------------------------------------
#
#   Product Analysis (request Pavel)
#           #TODO:update the functions (get,update,set)
# -----------------------------------------------------------------------------------------------------------

# --------------------------------------------------------------------------------------------------------------------------------------------
#
#   Assembly Sequence
#           #TODO: assembly contains lots of components without an ID or name (e.g. orientation, geometry), can we adapt this to enable re-use?
#           #TODO: hasFastner links to productPart, is that correct?
# ---------------------------------------------------------------------------------------------------------------------------------------------
    def getAssemblySequence(self,AssemblySystemName):
        """
         Function to fetch the Assembly sequence using the name of the assemblySystem as identifier.
         The function returns an interface block of the AssemblySequence

        :param string AssemblySystemName: Name identifier of the assemblySystem

         :return object AssemblySequenceModel: Interface object of the AssemblySequence model (-1 = error)
        """
        if len(self.model.includesAssemblySystem.items) != 0:
            for AssemblySystem in self.model.includesAssemblySystem.items:
                if AssemblySystem.hasName == AssemblySystemName:
                    try:
                        s_temp = AssemblySystem.hasAssemblySequence
                        #assemblyMetric
                        assemblyMetric = [s_temp.hasAssemblyMetric.items[0].hasName,s_temp.hasAssemblyMetric.items[0].hasValue]
                        #assemblyOption - recursive
                        assemblyOptions = self.getAssemblyOption(s_temp.hasAssemblyOption)

                        sequence = AssemblySequence(Name=assemblyOptions.Name, Description="NOT IN MM",AssemblyMetric=assemblyMetric,AssemblyOptions=assemblyOptions)

                        return sequence

                    except:
                        return -1
                else:
                    return -1
        else:
            return -1


    def updateAssemblySequence(self,AssemblySystemName,IntefaceObject):
        """
              Function to update the AssemblySequence using the name of the assemblySystem as identifier
              The AssemblySequence interface object is used to interface with the function.
              The function returns whether or not the function is performed correctly

              :param string AssemblySystemName: Name identifier of the assemblySystem
              :param object AssemblySequenceModel: Interface object of the AssemblySequence model

              :return int Error: -1 = error, 1= function performed correcly
        """
        #TODO: discuss if this is required
        x=10
        return -1


    def setAssemblySequence(self,AssemblySystemName,InterfaceObject):
        """
             Function to add a AssemblySequence model to an existing AssemblySystem using the name of the assemblySystem as identifier.
             The AssemblySequence interface object is used to interface with the function.
             The function returns whether or not the function is performed correctly

             :param string AssemblySystemName: Name identifier of the assemblySystem
              :param object AssemblySequenceModel: Interface object of the AssemblySequence model

             :return int Error: -1 = error, 1= function performed correcly
        """
        #-----------------------------------------------------------------------------------------------------
        #   DEBUG PURPOSE: Add product and assemblySystem if not exist
        #-----------------------------------------------------------------------------------------------------
        if len(self.model.includesAssemblySystem.items)==0:

            # ---------------------- PRODUCT DUMMY ---------------------------
            product = self.create_connected("Product")
            product.hasProductName = "DUMMY-PRODUCT-DEBUG"
            product.hasSTEPfile = 'c:/NEW/PATH/TO/THE/STEPFILE'

            # ---------------------- ASSEMBLY SYSTEM DUMMY ---------------------------
            asys = self.create_connected("AssemblySystem")

            # -- minimal dependecies --
            asys.hasName = AssemblySystemName
            asys.hasGUID = ""
            GravityDirection = self.create_connected("Direction")
            GravityDirection.hasRx=1.0
            GravityDirection.hasRy = 2.0
            GravityDirection.hasRz = 3.0
            GravityDirection.hasX = 4.0
            GravityDirection.hasY = 5.0
            GravityDirection.hasZ = 6.0
            asys.hasGravityDirection = GravityDirection

            GroundPosition = self.create_connected("Position")
            GroundPosition.hasX = 0.1
            GroundPosition.hasY = 0.2
            GroundPosition.hasZ = 0.3
            asys.hasGroundPosition = GroundPosition

            product.hasAssemblySystem.append(asys)

        else:
            # ---------------------- PRODUCT REAL ---------------------------
            product = self.model.includesProduct.items[0]
            # ---------------------- ASSEMBLY SYSTEM REAL ---------------------------

            # Search for corresponding assemblySystem and add - ASSUMPTION: everything else is set for the assemblySystem
            for a in self.model.includesAssemblySystem.items:
                if a.hasName == AssemblySystemName:
                    asys = a
        # -----------------------------------------------------------------------------------------------------
        #   Create assembly sequence
        # -----------------------------------------------------------------------------------------------------
        AssemblySequence = self.create_connected("AssemblySequence")

        #----add AssemblyMetric-----
        asm = self.create_connected("AssemblyMetric")
        asm.hasName=InterfaceObject.AssemblyMetric[0]
        asm.hasValue = InterfaceObject.AssemblyMetric[1]
        AssemblySequence.hasAssemblyMetric.append(asm)

        # -- recursive AssemblyOptions --
        AssemblySequence.hasAssemblyOption = self.setAssemblyOption_NEW(Option=InterfaceObject.AssemblyOptions)


        asys.hasAssemblySequence = AssemblySequence
        product.hasAssemblySystem.append(asys)

        self.updateKBv6()

        return -1


    def getAssemblyOption(self,Option):
        """
         Helper function to fetch the Assembly options for each Assembly step .
         The function returns a recursive AssemblyOption List

        :param object Option: Assembly option ecore object

         :return List AssemblyOptionList: list of assembly options RECURSIVE (-1 = error)
        """
        AssemblyOptionList = []
        assembly = None
        try:
            # operations
            operationList = []
            for operation in Option.hasOperation.items:
                op = Operation(Name=operation.hasName, Description="NOT IN MM")
                op.OperationMetric = [operation.hasOperationMetric.items[0].hasName, operation.hasOperationMetric.items[0].hasValue]  # TODO: in MM, multiple metrics can be added -> is this necessary?
                # operators
                operatorList = []
                for operator in operation.hasOperator.items:
                    operator_temp = Operator(Name=operator.hasName,CostPerHour=operator.hasCostPerHour,Height=operator.hasHeight,MaxLiftingWeight=operator.hasMaxLiftingWeight,Reach=operator.hasReach)
                    operatorList.append(operator_temp)
                op.Operators = operatorList
                # Tools
                ToolList = []
                for tool in operation.hasTool.items:
                    requiredOperators = []
                    for r_operator in operation.hasOperator.items:
                        requiredOperators.append(r_operator.hasName)  # TODO: only name reference added, ok?
                    Tool_temp = Tool(Name=tool.hasName,Mass=tool.hasMass,Cost=tool.hasCost,Orientation=tool.hasOrientation.hasData,Geometry=tool.hasGeometry.hasDataInStepFile,RequiredOperators=requiredOperators)
                    #Tool_temp = [tool.hasName, tool.hasMass, tool.hasCost, tool.hasOrientation.hasData,tool.hasGeometry.hasDataInStepFile, requiredOperators]
                    ToolList.append(Tool_temp)
                op.Tool = ToolList
                # Fastners
                FastnersList = []
                for fastner in operation.hasFastner.items:
                    FastnersList.append(fastner.hasPartName)  # TODO: only name reference added, ok?
                op.Fastners = FastnersList
                operationList.append(op)

            # -- define the assemblyOption --
            ASO = AssemblyOption(Name=Option.hasName,Operations=operationList)

            #Linking the recursive operations!
            subAssemblyOptions = []
            for recursive_option in Option.hasRecursiveAssemblyOption:
                subAssemblyOptions.append(self.getAssemblyOption(recursive_option))

            ASO.RecursiveAssemblyOption = subAssemblyOptions

            return ASO

        except:
            return []


    def setAssemblyOption_NEW(self, Option):
        try:

            # -- define the assemblyOption --
            ASO = self.create_connected("SubAssembly")
            ASO.hasName = Option.Name

            # operations
            for operation in Option.Operations:
                op = self.create_connected("Operation")

                # ----add OperationMetric-----
                om = self.create_connected("OperationMetric")
                om.hasName = operation.OperationMetric[0]
                om.hasValue = operation.OperationMetric[1]
                op.hasOperationMetric.append(om)

                # ----add Fastners----
                # TODO: productParts by name?

                # ----add Operators-----
                # TODO:implement after hacks are removed!

                # ----add Tools-----
                for tool in operation.Tool:
                    t = self.create_connected("Tool")
                    t.hasName = tool.Name
                    t.hasCost = tool.Cost
                    t.hasMass = tool.Mass

                    # --# ----Geometry-----
                    g = self.create_connected("Geometry")
                    g.hasDataInStepFile = tool.Geometry
                    t.hasGeometry = g
                    # --# ----Orientation-----
                    o = self.create_connected("Orientation")
                    o.hasData = tool.Orientation
                    t.hasOrientation = o

                    op.hasTool.append(t)

                    # -- Add operation to subassembly --
                    ASO.hasOperation.append(op)


            #Linking the recursive operations!
            subAssemblyOptions = []
            for recursive_option in Option.RecursiveAssemblyOption:
                tmp = self.setAssemblyOption_NEW(recursive_option)
                ASO.hasRecursiveAssemblyOption.append(tmp)

            #ASO.RecursiveAssemblyOption = subAssemblyOptions

            return ASO

        except:
            return []

# -----------------------------------------------------------------------------------------------------------
#
#   Contract-Ontology extension
#
# -----------------------------------------------------------------------------------------------------------
    def getContract(self,ContractName):
        """
          Function to fetch a contract model using the name as identifier.
          The function returns an interface block of the contract model

          :param string ContractName: Name identifier of the Contract

          :return object ContractModel: Interface object of the contract model (-1 = error)
        """
        return getContract(ContractName,self.model)

    def updateContract(self,interfaceObject):
        """
              Function to update a contract matching the name as identifier.
              The contract interface object is used to interface with the function.
              The function returns whether or not the function is performed correctly

              :param object ContractModel: Interface object of the Contract model

              :return int Error: -1 = error, 1= function performed correcly
            """
        return updateContract(interfaceObject, self.model, self.KBPath)

    def setContract(self,interfaceObject):
        """
                 Function to add a contract to an blank KB.
                 The contract interface object is used to interface with the function.
                 The function returns whether or not the function is performed correctly

                 :param object ContractModel: Interface object of the Contract model

                 :return int Error: -1 = error, 1= function performed correcly
        """
        return setContract(interfaceObject, self.model, self.KBPath)

    def getOntology(self):
        """
          Function to fetch the complete ontology model.
          The function returns an interface block of the ontology model

          :return object OntologyModel: Interface object of the ontology model (-1 = error)
        """

        return getOntology(self.model)

    def updateOntology(self,interfaceObject):
        """
              Function to update the complete ontology model.
              The ontology interface object is used to interface with the function.
              The function returns whether or not the function is performed correctly

              :param object OntologyModel: Interface object of the ontology model

              :return int Error: -1 = error, 1= function performed correcly
            """

        error = -1
        return updateOntology(interfaceObject, self.model, self.KBPath)

    def setOntology(self,interfaceObject):
        """
                 Function to add the ontology model to a blank KB.
                 The ontology interface object is used to interface with the function.
                 The function returns whether or not the function is performed correctly

                 :param object OntologyModel: Interface object of the ontology model

                 :return int Error: -1 = error, 1= function performed correcly
        """
        error = -1
        return setOntology(interfaceObject, self.model, self.KBPath)








# -----------------------------------------------------------------------------------------------------------
#
#   HELPER FUNCTIONS
#
# -----------------------------------------------------------------------------------------------------------

    def updateKB(self):
        """
              For internal use only!
            """
        updateKB(self.model)

    def updateKBv6(self):
        """
              For internal use only!
            """
        # TODO: update the current MM instead of the new one
        updateKBv6(self.KB_path, self.model)

    def resolvePath(self, relativePath):
        """
              For internal use only!
            """
        return resolvePath(relativePath)

    def importInstanceModel(self, path_ecore, path_instance, VERBOSE=False):
        """
              For internal use only!
            """
        self.model = importInstanceModel(path_ecore, path_instance, VERBOSE=False)
        return self.model

    def importInstanceModel_NEW(self, path_ecore, path_instance, VERBOSE=False):
        """
              For internal use only!
            """
        self.MM,self.model,self.model_instance = importInstanceModel_NEW(path_ecore, path_instance, VERBOSE=False)
        return self.MM,self.model,self.model_instance

    def addXMLResource(self, rset):
        """
              For internal use only!
            """
        return addXMLResource(rset)

    def initialize(self, model_uri):
        """
              For internal use only!
            """
        self.MM, self.model_instance = initialize(model_uri)

    def serialize(self, xmi_uri):
        """
              For internal use only!
            """
        serialize(xmi_uri, self.model)

    def create(self, e_class):
        """
              For internal use only!
            """
        return create(e_class, self.MM, self.model)

    def create_connected(self, e_class):
        """
              For internal use only!
            """
        return create_connected(e_class, self.MM, self.model)

    def create_custom(self, parent, e_class):
        """
              For internal use only!
            """
        return create_custom(parent, e_class, self.MM)

    def create_noPlatformRoot(self, e_class):
        """
              For internal use only!
            """
        return create_noPlatformRoot(e_class, self.MM, self.model)

    # export the meta model to a file
    def export_metamodel(self, output_file):
        """
              For internal use only!
            """
        export_metamodel(output_file, self.model)


# -----------------------------------------------------------------------------------------------------------
#
#   Possible usefull input/code
#
# -----------------------------------------------------------------------------------------------------------

    # -------------------------------------------------------------------------------------------------------------
    # PAVEL INPUT ==>
    #            - ModelFile String
    #            - AnalysisType String
    #            - substepNum Integer
    #            - MeshType [string1,string2]
    #            -boundaryconditions    [bc1,bc2]
    #            -Loads    [Load1,Load2]
    #            -Contacts    [c1,c2]
    # --------------------------------------------------------------------------------------------------------------
    def setProductAnalysis(self, AnalysisList):
        if self.model.includesAnalysis.items != 0:
            for an_new in AnalysisList:
                found = 0
                for an_old in self.model.includesAnalysis.items:
                    if an_new[1] == an_old.hasAnalysisType:  # TODO: for now matched on the type but better on name?
                        found = 1
                        print("Performance Model already added - Do we need to update them?")
                        old = an_old
                        new = an_new

                if found == 0:  # no matching productPart found -> add?
                    print("Performance Model not found - Do we need to add them?")
                    part = self.createAnalysis(an_new)
                    self.model.includesAnalysis.items.append(part)

        else:
            for an_new in AnalysisList:  # assuming that no performance models are added yet
                part = self.createAnalysis(an_new)
                self.model.includesAnalysis.items.append(part)
        self.updateKB()

    # -------------------------------------------
    #
    #       HELPER FUNCTION TO CREATE PRODUCT PART
    # -------------------------------------------
    def createAnalysis(self, analysis):
        part = None

        an_new = self.create("Analysis")
        # an_new.hasName = analysis[0]   #TODO: not in current MM
        an_new.hasAnalysisType = analysis[1]

        # add model
        # model = self.create('AnalysisModel')
        # model.hasName = ""                  #TODO:not yet in API call signature
        # model.hasDescription = ""  # TODO:not yet in API call signature
        # model.hasVersion = ""  # TODO:not yet in API call signature
        # model.hasFile = analysis[2]  # TODO:not yet MM
        # an_new.hasAnalysisModel = model

        # add optional boundary conditions
        if len(analysis[5]) != 0:
            x = 0
        else:
            print("No Boundary Conditions specified")

        # add optional loads
        if len(analysis[6]) != 0:
            for l in analysis[6]:
                load = self.create('Load')
                load.hasMagnitude = l[1]
                load.hasLoadType = l[0]
                # applicationArea = self.create('ApplicationArea')    #TODO: this is underspecified in the API call => needs reference to productPart, contactFeature,...

                an_new.hasLoad.items.append(load)
        else:
            print("No Loads specified")

        # add optional contacts
        if len(analysis[7]) != 0:
            x = 0
        else:
            print("No Contacts specified")

        return an_new

    # -------------------------------------------------------------------------------------------------------------
    # PAVEL INPUT ==>
    #            - add model parameters (helper function)
    #            - parameter as [hasName,hasDescription,hasKey,hasValue]
    # --------------------------------------------------------------------------------------------------------------
    def addModelParameters(self, modelName, modelType, parameterList):

        model = self.create("Model")
        model.hasName = modelName
        model.hasModelType = modelType
        for par in parameterList:
            p = self.create("Parameter")
            p.hasName = par[0]
            p.hasDescription = par[1]
            p.hasKey = par[2]
            p.hasValue = par[3]
            model.hasModelParameter.items.append(p)

        return model



# ----------------------------------------------------------------------------------------------------------------------------------------------------
#                                                 SECONDARY FUNCTIONS
#                       TODO: discuss which of the secondary functions is required
# ----------------------------------------------------------------------------------------------------------------------------------------------------

    # # ---------------------------------node: Product attribute: Name--------------------------------------------------------------------
#     # def getProductName(self):
#     #     """
#     #      Function to fetch the Product name.
#     #
#     #
#     #      :return string Name: Product name (-1 = error)
#     #     """
#     #     if len(self.model.includesProduct.items) != 0:
#     #         value = self.model.includesProduct.items[0].hasProductName
#     #         return value
#     #     else:
#     #         return -1  # no Product instance
#     #
#     # def setProductName(self, att):
#     #     """
#     #      Function to set the Product name.
#     #
#     #      :param string name: Product name
#     #     """
#     #     self.model.includesProduct.items[0].hasProductName = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     #
#     # # ---------------------------------node: Product attribute: STEPfile--------------------------------------------------------------------
#     # def getProductSTEPfile(self):
#     #     """
#     #      Function to fetch the Product STEP file.
#     #
#     #      :return string STEPfile: path to the product STEP file (-1 = error)
#     #     """
#     #     if len(self.model.includesProduct.items) != 0:
#     #         value = self.model.includesProduct.items[0].hasSTEPfile  # TODO:update according to MM
#     #         return value
#     #     else:
#     #         return -1  # no Product instance
#     #
#     # def setProductSTEPfile(self, att):
#     #     """
#     #         Function to set the STEP file
#     #     :param att: path to the STEP file
#     #     """
#     #     self.model.includesProduct.items[0].hasSTEPfile = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: GroundPosition attribute: X--------------------------------------------------------------------
#     # def getGroundPositionX(self):
#     #     """
#     #      Function to fetch the GroundPositionX.
#     #
#     #
#     #      :return float GroundPositionX: GroundPositionX (-1 = error)
#     #     """
#     #     if len(self.model.includesAssemblySystem.items) != 0:
#     #         value = self.model.includesAssemblySystem.items[0].hasGroundPosition.hasX  # TODO:update according to MM
#     #         return value
#     #     else:
#     #         return -1  # no GroundPosition instance
#     #
#     # def setGroundPositionX(self, att):
#     #     """
#     #         Function to set the ground position X
#     #     :param float GroundPositionX: GroundPositionX
#     #     """
#     #     self.model.includesAssemblySystem.items[0].hasGroundPosition.hasX = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: GroundPosition attribute: Y--------------------------------------------------------------------
#     # def getGroundPositionY(self):
#     #     """
#     #      Function to fetch the GroundPositionY.
#     #
#     #
#     #      :return float GroundPositionY: GroundPositionY (-1 = error)
#     #     """
#     #     if len(self.model.includesAssemblySystem.items) != 0:
#     #         value = self.model.includesAssemblySystem.items[0].hasGroundPosition.hasY  # TODO:update according to MM
#     #         return value
#     #     else:
#     #         return -2  # no GroundPosition instance
#     #
#     # def setGroundPositionY(self, att):
#     #     """
#     #         Function to set the ground position Y
#     #     :param float GroundPositionY: GroundPositionY
#     #     """
#     #     self.model.includesAssemblySystem.items[0].hasGroundPosition.hasY = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: GroundPosition attribute: Z--------------------------------------------------------------------
#     # def getGroundPositionZ(self):
#     #     """
#     #      Function to fetch the GroundPositionZ.
#     #
#     #
#     #      :return float GroundPositionZ: GroundPositionZ (-1 = error)
#     #     """
#     #     if len(self.model.includesAssemblySystem.items) != 0:
#     #         value = self.model.includesAssemblySystem.items[0].hasGroundPosition.hasZ  # TODO:update according to MM
#     #         return value
#     #     else:
#     #         return -2  # no GroundPosition instance
#     #
#     # def setGroundPositionZ(self, att):
#     #     """
#     #         Function to set the ground position Z
#     #         :param float GroundPositionZ: GroundPositionZ
#     #     """
#     #     self.model.includesAssemblySystem.items[0].hasGroundPosition.hasZ = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: GravityDirection attribute: X--------------------------------------------------------------------
#     # def getGravityDirectionX(self):
#     #     """
#     #      Function to fetch the GravityDirectionX.
#     #
#     #
#     #      :return float GravityDirectionX: GravityDirectionX (-1 = error)
#     #     """
#     #     if len(self.model.includesAssemblySystem.items) != 0:
#     #         value = self.model.includesAssemblySystem.items[0].hasGravityDirection.hasX  # TODO:update according to MM
#     #         return value
#     #     else:
#     #         return -2  # no GravityDirection instance
#     #
#     # def setGravityDirectionX(self, att):
#     #     """
#     #         Function to set the gravitation position X
#     #     :param float GravityDirectionX: GravityDirectionX
#     #     """
#     #     self.model.includesAssemblySystem.items[0].hasGravityDirection.hasX = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: GravityDirection attribute: Y--------------------------------------------------------------------
#     # def getGravityDirectionY(self):
#     #     """
#     #      Function to fetch the GravityDirectionY.
#     #
#     #
#     #      :return float GravityDirectionY: GravityDirectionY (-1 = error)
#     #     """
#     #     if len(self.model.includesAssemblySystem.items) != 0:
#     #         value = self.model.includesAssemblySystem.items[0].hasGravityDirection.hasY  # TODO:update according to MM
#     #         return value
#     #     else:
#     #         return -2  # no GravityDirection instance
#     #
#     # def setGravityDirectionY(self, att):
#     #     """
#     #         Function to set the gravitation position Y
#     #     :param float GravityDirectionY: GravityDirectionY
#     #     """
#     #     self.model.includesAssemblySystem.items[0].hasGravityDirection.hasY = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: GravityDirection attribute: Z--------------------------------------------------------------------
#     # def getGravityDirectionZ(self):
#     #     """
#     #      Function to fetch the GravityDirectionZ.
#     #
#     #
#     #      :return float GravityDirectionZ: GravityDirectionZ (-1 = error)
#     #     """
#     #     if len(self.model.includesAssemblySystem.items) != 0:
#     #         value = self.model.includesAssemblySystem.items[0].hasGravityDirection.hasZ  # TODO:update according to MM
#     #         return value
#     #     else:
#     #         return -2  # no GravityDirection instance
#     #
#     # def setGravityDirectionZ(self, att):
#     #     """
#     #         Function to set the gravitation position Z
#     #     :param float GravityDirectionZ: GravityDirectionZ
#     #     """
#     #     self.model.includesAssemblySystem.items[0].hasGravityDirection.hasZ = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     #
#     # # ---------------------------------node: ProductPart attribute: name--------------------------------------------------------------------
#     # def setProductPartname(self,name, att):
#     #     for productPart in self.model.includesProductPart.items:
#     #         if productPart.hasPartName == name:
#     #             productPart.hasPartName = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: ProductPart attribute: type--------------------------------------------------------------------
#     # def getProductPartType(self, name):
#     #     """
#     #      Function to fetch the ProductPartType.
#     #
#     #
#     #      :return string ProductPartType: ProductPartType (-1 = error)
#     #     """
#     #     value = None
#     #     for productPart in self.model.includesProductPart.items:
#     #         if productPart.hasPartName == name:
#     #             value = productPart.hasPartType  # TODO:update according to MM
#     #     return value
#     #
#     # def setProductParttype(self,name, att):
#     #     for productPart in self.model.includesProductPart.items:
#     #         if productPart.hasPartName == name:
#     #             productPart.hasPartType = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: Operator attribute: name--------------------------------------------------------------------
#     # def getOperatorname(self):
#     #     """
#     #      Function to fetch the Operatorname.
#     #
#     #
#     #      :return string Operatorname: Operatorname (-1 = error)
#     #     """
#     #     value = None
#     #     if len(self.model.includesOperator.items) != 0:
#     #         value = self.model.includesOperator.items[0].hasname  # TODO:update according to MM
#     #     return value
#     #
#     #
#     # def setOperatorname(self, att):
#     #     self.model.includesOperator.items[0].hasname = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: Operator attribute: reach--------------------------------------------------------------------
#     # def getOperatorreach(self):
#     #     """
#     #      Function to fetch the Operatorreach.
#     #
#     #
#     #      :return float Operatorreach: Operatorreach (-1 = error)
#     #     """
#     #     if len(self.model.includesOperator.items) != 0:
#     #         value = self.model.includesOperator.items[0].hasreach  # TODO:update according to MM
#     #         return value
#     #     else:
#     #         return -2  # no Operator instance
#     #
#     # def setOperatorreach(self, att):
#     #     self.model.includesOperator.items[0].hasreach = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: Operator attribute: height--------------------------------------------------------------------
#     # def getOperatorheight(self):
#     #     """
#     #      Function to fetch the Operatorheight.
#     #
#     #
#     #      :return float Operatorheight: Operatorheight (-1 = error)
#     #     """
#     #     if len(self.model.includesOperator.items) != 0:
#     #         value = self.model.includesOperator.items[0].hasheight  # TODO:update according to MM
#     #         return value
#     #     else:
#     #         return -2  # no Operator instance
#     #
#     # def setOperatorheight(self, att):
#     #     self.model.includesOperator.items[0].hasheight = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: Operator attribute: maxLiftingWeight--------------------------------------------------------------------
#     # def getOperatormaxLiftingWeight(self):
#     #     """
#     #      Function to fetch the OperatormaxLiftingWeight.
#     #
#     #
#     #      :return float OperatormaxLiftingWeight: OperatormaxLiftingWeight (-1 = error)
#     #     """
#     #     if len(self.model.includesOperator.items) != 0:
#     #         value = self.model.includesOperator.items[0].hasmaxLiftingWeight  # TODO:update according to MM
#     #         return value
#     #     else:
#     #         return -2  # no Operator instance
#     #
#     # def setOperatormaxLiftingWeight(self, att):
#     #     self.model.includesOperator.items[0].hasmaxLiftingWeight = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: Operator attribute: costPerHour--------------------------------------------------------------------
#     # def getOperators(self):
#     #     operatorList = []
#     #     if len(self.model.includesOperator.items) != 0:
#     #         for operator in self.model.includesOperator.items:
#     #             operatorList.append([operator.hasName,operator.reach,operator.height,operator.maxLiftingWeight,operator.costPerHour])
#     #     return operatorList
#     #
#     # def setOperators(self, list):
#     #     for op in list:       #list = [operator1,operator2,...,operatorn] with operator = [name,reach,height,maxLiftingWeight,costPerHour]
#     #         #create and add operator
#     #         operator = self.create("Operator")
#     #         operator.hasName = op[0]
#     #         operator.reach = op[1]
#     #         operator.height = op[2]
#     #         operator.maxLiftingWeight = op[3]
#     #         operator.costPerHour = op[4]
#     #         self.model.includesOperator.items.append(operator)
#     #
#     #     self.updateKB()
#     #
#     #
#     # # ---------------------------------node: Analysis attribute: analysisFile--------------------------------------------------------------------
#     # def getAnalysis(self):
#     #     if len(self.model.includesAnalysis.items) != 0:
#     #         analysisList = []
#     #         for analysis in self.model.includesAnalysis.items:
#     #             analysisList.append([analysis.hasAnalysisType])  # TODO:update according to MM
#     #         return analysisList
#     #     else:
#     #         return -2  # no ProductPart instance
#     #
#     # def getAnalysisFile(self,type):
#     #     for analysis in self.model.includesAnalysis.items:
#     #         if analysis.hasAnalysisType == type:
#     #             return analysis.hasAnalysisFile   # TODO:update according to MM
#     #
#     # def setAnalysisanalysisFile(self,type, att):
#     #     for analysis in self.model.includesAnalysis.items:
#     #         if analysis.hasAnalysisType == type:
#     #             analysis.hasAnalysisFile = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: Analysis attribute: type--------------------------------------------------------------------
#     # def getAnalysistype(self,type):
#     #     for analysis in self.model.includesAnalysis.items:
#     #         if analysis.hasAnalysisType == type:
#     #             return analysis.hasAnalysisType  # TODO:update according to MM
#     #
#     # def setAnalysistype(self,type, att):
#     #     for analysis in self.model.includesAnalysis.items:
#     #         if analysis.hasAnalysisType == type:
#     #             analysis.hasAnalysisType =att
#     #     self.updateKB()
#     #
#     # def getAnalysisResponsePoints(self,type):
#     #     responsePointList = []
#     #     for analysis in self.model.includesAnalysis.items:
#     #         if analysis.hasAnalysisType == type:
#     #             for responsPoint in analysis.hasResponsePoint.items:
#     #                 x = responsPoint.hasPosition.hasX
#     #                 y = responsPoint.hasPosition.hasY
#     #                 z = responsPoint.hasPosition.hasZ
#     #                 responsePointList.append([responsPoint.hasResponseType,[x,y,z]])
#     #     return responsePointList
#     #
#     # def setAnalysisResponsePoints(self,type, list):     #TODO: responspoints = [point1,point2,...,pointn] with point[type,X,Y,Z]
#     #     for analysis in self.model.includesAnalysis.items:
#     #         if analysis.hasAnalysisType == type:
#     #             for responsepoint in list:
#     #                 #create and add position
#     #                 position = self.create("Position")
#     #                 position.hasX = responsepoint[1]
#     #                 position.hasY = responsepoint[2]
#     #                 position.hasZ = responsepoint[3]
#     #                 #add responsePoint
#     #                 point = self.create("ResponsePoint")
#     #                 point.hasResponseType = responsepoint[0]
#     #                 point.hasPosition = position
#     #                 analysis.hasResponsePoint.items.append(point)
#     #             x = 0
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: Analysis attribute: BoundaryCondition--------------------------------------------------------------------
#     # def getAnalysisBoundaryCondition(self,analysisID):
#     #     value = None
#     #     productPart = None
#     #     contactFeatures = []
#     #     blockDOF = []
#     #     if len(self.model.includesAnalysis.items) != 0:
#     #         for analysis in self.model.includesAnalysis.items:
#     #             if analysis.hasAnalysisType == analysisID:  #TODO: ADAPT THIS TO THE ID
#     #                 #get the applicationArea - contact features
#     #                 if len(analysis.hasBoundaryCondition.items[0].hasApplicationArea.hasContactFeature.items) != 0:
#     #                     for feature in analysis.hasBoundaryCondition.items[0].hasApplicationArea.hasContactFeature.items:
#     #                         contactFeatures.append(feature.hasFeatureStepFile)
#     #                 # get the applicationArea - product parts
#     #                     productPart = analysis.hasBoundaryCondition.items[0].hasApplicationArea.hasProductPart.hasPartName
#     #                 #get the blockDOFs
#     #                 if len(analysis.hasBoundaryCondition.items)!= 0:
#     #                     blockDOF = [analysis.hasBoundaryCondition.items[0].hasBlockedDOF.DOF1,analysis.hasBoundaryCondition.items[0].hasBlockedDOF.DOF2,analysis.hasBoundaryCondition.items[0].hasBlockedDOF.DOF3,analysis.hasBoundaryCondition.items[0].hasBlockedDOF.DOF4,analysis.hasBoundaryCondition.items[0].hasBlockedDOF.DOF5,analysis.hasBoundaryCondition.items[0].hasBlockedDOF.DOF6]
#     #
#     #
#     #         return productPart,contactFeatures,blockDOF
#     #     else:
#     #         return -2  # no Analysis instance
#     #
#     #
#     # def setAnalysisBoundaryCondition(self, att):
#     #     self.model.includesAnalysis.items[0].hasBoundaryCondition = att  # TODO:update according to MM
#     #
#     #
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: AnalysisParameter attribute: type--------------------------------------------------------------------
#     # def getAnalysisParameters(self,analysisID):
#     #     parameterList = []
#     #     if len(self.model.includesAnalysis.items) != 0:
#     #         for analysis in self.model.includesAnalysis.items:
#     #             if analysis.hasAnalysisType == analysisID:  # TODO: ADAPT THIS TO THE ID
#     #                 if len(analysis.hasAnalysisParameter.items) != 0:
#     #                     for parameter in analysis.hasAnalysisParameter.items:
#     #                         key = parameter.hasKey
#     #                         value = parameter.hasValue
#     #                         parameterList.append([key,value])
#     #
#     #
#     #     return parameterList
#     #
#     # def setAnalysisParametertype(self, att):
#     #     self.model.includesAnalysisParameter.items[0].hastype = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: AnalysisParameter attribute: name--------------------------------------------------------------------
#     # def getAnalysisParametername(self):
#     #     if len(self.model.includesAnalysisParameter.items) != 0:
#     #         value = self.model.includesAnalysisParameter.items[0].hasname  # TODO:update according to MM
#     #         return value
#     #     else:
#     #         return -2  # no AnalysisParameter instance
#     #
#     # def setAnalysisParametername(self, att):
#     #     self.model.includesAnalysisParameter.items[0].hasname = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: AnalysisParameter attribute: value--------------------------------------------------------------------
#     # def getAnalysisParametervalue(self):
#     #     if len(self.model.includesAnalysisParameter.items) != 0:
#     #         value = self.model.includesAnalysisParameter.items[0].hasvalue  # TODO:update according to MM
#     #         return value
#     #     else:
#     #         return -2  # no AnalysisParameter instance
#     #
#     # def setAnalysisParametervalue(self, att):
#     #     self.model.includesAnalysisParameter.items[0].hasvalue = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: Parameter attribute: parameterEffect--------------------------------------------------------------------
#     # def getParameterparameterEffect(self):
#     #     if len(self.model.includesParameter.items) != 0:
#     #         value = self.model.includesParameter.items[0].hasparameterEffect  # TODO:update according to MM
#     #         return value
#     #     else:
#     #         return -2  # no Parameter instance
#     #
#     # def setParameterparameterEffect(self, att):
#     #     self.model.includesParameter.items[0].hasparameterEffect = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: Load attribute: type--------------------------------------------------------------------
#     # def getLoad(self,analysisID):
#     #     LoadList = []
#     #     contactFeatures = []
#     #     productPart = None
#     #     direction = []
#     #     LoadType = None
#     #     LoadMagnitude = 0.0
#     #     if len(self.model.includesAnalysis.items) != 0:
#     #         for analysis in self.model.includesAnalysis.items:
#     #             if analysis.hasAnalysisType == analysisID:  # TODO: ADAPT THIS TO THE ID
#     #                 # get the applicationArea - contact features
#     #                 if len(analysis.hasLoad.items[0].hasApplicationArea.hasContactFeature.items) != 0:
#     #                     for feature in analysis.hasBoundaryCondition.items[
#     #                         0].hasApplicationArea.hasContactFeature.items:
#     #                         contactFeatures.append(feature.hasFeatureStepFile)
#     #                 # get the applicationArea - product parts
#     #                 productPart = analysis.hasLoad.items[0].hasApplicationArea.hasProductPart.hasPartName
#     #                 # get load attributes
#     #                 if len(analysis.hasLoad.items) != 0:
#     #                     for load in analysis.hasLoad.items:
#     #                         LoadType = load.hasLoadType
#     #                         LoadMagnitude = load.hasMagnitude
#     #                         # get load direction
#     #                         direction = [load.hasDirection.hasX,load.hasDirection.hasY,load.hasDirection.hasZ]
#     #                 LoadList.append([LoadType,LoadMagnitude,contactFeatures,productPart,direction])
#     #         return LoadList
#     #     else:
#     #         return -2  # no Load instance
#     #
#     # def setLoad(self,analysisID, att):
#     #     self.model.includesLoad.items[0].hastype = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: Tool attribute: name--------------------------------------------------------------------
#     # def getToolname(self):
#     #     if len(self.model.includesTool.items) != 0:
#     #         value = self.model.includesTool.items[0].hasname  # TODO:update according to MM
#     #         return value
#     #     else:
#     #         return -2  # no Tool instance
#     #
#     # def setToolname(self, att):
#     #     self.model.includesTool.items[0].hasname = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: Tool attribute: mass--------------------------------------------------------------------
#     # def getToolmass(self):
#     #     if len(self.model.includesTool.items) != 0:
#     #         value = self.model.includesTool.items[0].hasmass  # TODO:update according to MM
#     #         return value
#     #     else:
#     #         return -2  # no Tool instance
#     #
#     # def setToolmass(self, att):
#     #     self.model.includesTool.items[0].hasmass = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: Tool attribute: cost--------------------------------------------------------------------
#     # def getTools(self):
#     #     toolList = []
#     #     if len(self.model.includesTool.items) != 0:
#     #         for tool in self.model.includesTool.items:
#     #             toolList.append([tool.hasName,tool.hasCost,tool.hasMass,tool.hasOrientation.hasData,tool.hasGeometry.hasDataInStepFile])
#     #     return toolList
#     #
#     #
#     # def setTools(self, list):
#     #
#     #     for t in list:      #list with [tool1,tool2,...,tooln] with tool = [name,cost,mass,orientationData,[Operator],geomestryData] with operator = [name,reach,height,maxLift,costPerHour]
#     #         #create tool
#     #         tool = self.create("Tool")
#     #         tool.hasName = t[0]
#     #         tool.hasCost = t[1]
#     #         tool.hasMass = t[2]
#     #
#     #         #create orientation
#     #         orientation = self.create("Orientation")
#     #         orientation.hasData = t[3]
#     #         tool.hasOrientation = orientation
#     #
#     #         # create operator   #TODO==> do we need to search for existing operators??
#     #             #search for existing operators
#     #
#     #             #if no matching existing operator, make new
#     #         operator = self.create("Operator")
#     #         operator.hasName = t[4][0]
#     #         operator.reach = t[4][1]
#     #         operator.height = t[4][2]
#     #         operator.maxLiftingWeight = t[4][3]
#     #         operator.costPerHour = t[4][4]
#     #
#     #         # create geometry
#     #         geometry = self.create("Geometry")
#     #         geometry.hasDataInStepFile = t[5]
#     #         tool.hasGeometry = geometry
#     #         self.model.includesTool.items.append(tool)
#     #
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: OperationMetric attribute: name--------------------------------------------------------------------
#     # def getOperationMetricname(self):
#     #     if len(self.model.includesOperationMetric.items) != 0:
#     #         value = self.model.includesOperationMetric.items[0].hasName  # TODO:update according to MM
#     #         return value
#     #     else:
#     #         return -2  # no OperationMetric instance
#     #
#     # def setOperationMetricname(self, att):
#     #     self.model.includesOperationMetric.items[0].hasName = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: OperationMetric attribute: value--------------------------------------------------------------------
#     # def getOperationMetricvalue(self):
#     #     if len(self.model.includesOperationMetric.items) != 0:
#     #         value = self.model.includesOperationMetric.items[0].hasValue  # TODO:update according to MM
#     #         return value
#     #     else:
#     #         return -2  # no OperationMetric instance
#     #
#     # def setOperationMetricvalue(self, att):
#     #     self.model.includesOperationMetric.items[0].hasValue = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: AssemblyMetric attribute: value--------------------------------------------------------------------
#     # def getAssemblyMetricvalue(self):
#     #     if len(self.model.includesAssemblyMetric.items) != 0:
#     #         value = self.model.includesAssemblyMetric.items[0].hasvalue  # TODO:update according to MM
#     #         return value
#     #     else:
#     #         return -1  # no AssemblyMetric instance
#     #
#     # def setAssemblyMetricvalue(self, att):
#     #     self.model.includesAssemblyMetric.items[0].hasvalue = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: Material attribute: name--------------------------------------------------------------------
#     #
#     # def getMaterialname(self, partName):
#     #     value = None
#     #     for productPart in self.model.includesProductPart.items:
#     #         if productPart.hasPartName == partName:
#     #             value = productPart.hasMaterial.hasName  # TODO:update according to MM
#     #
#     #     return value
#     #
#     # def setMaterialname(self, partName, att):
#     #     for productPart in self.model.includesProductPart.items:
#     #         if productPart.hasPartName == partName:
#     #             productPart.hasMaterial.hasName = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     #     # ---------------------------------node: Material attribute: density--------------------------------------------------------------------
#     #
#     # def getMaterialdensity(self, partName):
#     #     value = None
#     #     for productPart in self.model.includesProductPart.items:
#     #         if productPart.hasPartName == partName:
#     #             value = productPart.hasMaterial.hasDensity  # TODO:update according to MM
#     #
#     #     return value
#     #
#     # def setMaterialdensity(self, partName, att):
#     #     for productPart in self.model.includesProductPart.items:
#     #         if productPart.hasPartName == partName:
#     #             productPart.hasMaterial.hasDensity = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     #     # ---------------------------------node: Material attribute: cost--------------------------------------------------------------------
#     #
#     # def getMaterialcost(self, partName):
#     #     value = None
#     #     for productPart in self.model.includesProductPart.items:
#     #         if productPart.hasPartName == partName:
#     #             value = productPart.hasMaterial.hasCost  # TODO:update according to MM
#     #
#     #     return value
#     #
#     # def setMaterialcost(self, partName, att):
#     #     for productPart in self.model.includesProductPart.items:
#     #         if productPart.hasPartName == partName:
#     #             productPart.hasMaterial.hasCost = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     #     # ---------------------------------node: Material attribute: damping--------------------------------------------------------------------
#     #
#     # def getMaterialdamping(self, partName):
#     #     value = None
#     #     for productPart in self.model.includesProductPart.items:
#     #         if productPart.hasPartName == partName:
#     #             value = productPart.hasMaterial.hasDamping  # TODO:update according to MM
#     #
#     #     return value
#     #
#     # def setMaterialdamping(self, partName, att):
#     #     for productPart in self.model.includesProductPart.items:
#     #         if productPart.hasPartName == partName:
#     #             productPart.hasMaterial.hasDamping = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     #     # ---------------------------------node: Material attribute: youngModulus--------------------------------------------------------------------
#     #
#     # def getMaterialyoungModulus(self, partName):
#     #     value = None
#     #     for productPart in self.model.includesProductPart.items:
#     #         if productPart.hasPartName == partName:
#     #             value = productPart.hasMaterial.hasYoungModulus  # TODO:update according to MM
#     #
#     #     return value
#     #
#     # def setMaterialyoungModulus(self, partName, att):
#     #     for productPart in self.model.includesProductPart.items:
#     #         if productPart.hasPartName == partName:
#     #             productPart.hasMaterial.hasYoungModulus = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     #     # ---------------------------------node: Material attribute: poissonRatio--------------------------------------------------------------------
#     #
#     # def getMaterialpoissonRatio(self, partName):
#     #     value = None
#     #     for productPart in self.model.includesProductPart.items:
#     #         if productPart.hasPartName == partName:
#     #             value = productPart.hasMaterial.hasPoissonRatio  # TODO:update according to MM
#     #
#     #     return value
#     #
#     # def setMaterialpoissonRatio(self, partName, att):
#     #     for productPart in self.model.includesProductPart.items:
#     #         if productPart.hasPartName == partName:
#     #             productPart.hasMaterial.hasPoissonRatio = att  # TODO:update according to MM
#     #     self.updateKB()
#     #
#     # # ---------------------------------node: Material attribute: transparency--------------------------------------------------------------------
#     #
#     # def getMaterialtransparency(self, partName):
#     #     value = None
#     #     for productPart in self.model.includesProductPart.items:
#     #         if productPart.hasPartName == partName:
#     #             value = productPart.hasMaterial.hasTransparency  # TODO:update according to MM
#     #
#     #     return value
#     #
#     # def setMaterialtransparency(self, partName, att):
#     #     for productPart in self.model.includesProductPart.items:
#     #         if productPart.hasPartName == partName:
#     #             productPart.hasMaterial.hasTransparency = att  # TODO:update according to MM
#     #     self.updateKB()













import json

#TODO: discuss these interface objects as there are the main API interfacing mechanisms

class ASG():
    """
        ASG: Class representing the Assembly Sequence Generation (ASG) algorithm

         :param string Name: Name of the ASG
         :param string Description: Description of the ASG
         :param enum ProcessingType: ProcessingType of the ASGASG
         :param list Generator: Generator description of the ASG
         :param list Selector: Selector description of the ASG
         :param list Evaluator: Evaluator description of the ASG
         :param list StopConditionList: StopConditionList description of the ASG
         :param objectlist SelectorDFARuleList: list of DFA rules linked to the ASG selector
         :param objectlist EvaluatorDFARuleList: list of DFA rules linked to the ASG evaluator
    """
    def __init__(self,JSONDescriptor=None,Name="",Description="",ProcessingType="Full",Generator=[],Selector=[],SelectorDFARules = [],Evaluator=[],EvaluatorDFARules = [],StopConditions=[],DEBUG=True):
        self.DEBUG = DEBUG
        if JSONDescriptor is None:
            self.Name = Name
            self.Description = Description
            self.ProcessingType = str(ProcessingType)   #TODO: ENUM not serializable!
            #ASG GENERATOR config
            self.Generator = Generator
            # ASG SELECTOR config
            self.Selector = Selector
            self.SelectorDFARuleList = SelectorDFARules
            # ASG EVALUATOR config
            self.Evaluator = Evaluator
            self.EvaluatorDFARuleList = EvaluatorDFARules
            # ASG Terminator config
            self.StopConditionList=StopConditions
        else:
            self.json2object(JSONDescriptor)
            x=10

    def json2object(self,jsonDescriptor):
        """
             Function to generate a ASG object from a JSON file

             :param string jsonDescriptor: absolute path to the json file for the DFA rule

        """
        #--interpret JSON file--
        with open(jsonDescriptor, "r") as read_file:
            jsonObject = json.load(read_file)
        #--setup object--
        self.Name = jsonObject['Name']
        self.Description = jsonObject['Description']
        self.ProcessingType = jsonObject['ProcessingType']
        self.Generator = jsonObject['Generator']
        self.Selector = jsonObject['Selector']
        self.Evaluator = jsonObject['Evaluator']
        self.Generator = jsonObject['Generator']
        self.StopConditionList = []
        for terminator in jsonObject['StopConditionList']:
            s = StopCondition(None, terminator['Name'], terminator['Description'],terminator['Value'], terminator['StopCriteria'])
            self.StopConditionList.append(s)
        #TODO: discuss if we need to add the rules at once?


    def object2json(self):
        """
            Function to generate a json file from the ASG
        """
        return json.dumps(self, default=lambda o: o.__dict__,indent=4)

    def object2yaml(self):
        """
            Function to generate a yaml file from the ASG - PROTOTYPE
        """
        return -1 #yaml.dump(self,indent=4)

class DFARule():
    """
        DFARule: Class representing the DFA rule

         :param string Name: Name of the DFA rule
         :param string Description: Description of the DFA rule
         :param string RuleType: RuleType of the DFA rule
         :param bool isAppliedToProduct: Identify if DFA rule is applied to product
         :param bool isAppliedToProductPart: Identify if DFA rule is applied to product part
         :param bool isAppliedToAssemblySequence: Identify if DFA rule is applied to assembly sequence
         :param enum hasScorePropagation: hasScorePropagation of the DFA rule
         :param enum hasScoreType: hasScoreType of the DFA rule
         :param objectlist optionList: List of options of the DFA rule
         :param object property: Referenced property of the DFA rule
         :param list preprocessedData: preprocessedData of the DFA rule

    """
    def __init__(self, JSONDescriptor=None, Name="", Description="", RuleType='',isAppliedToProduct=False, isAppliedToProductPart=False,isAppliedToAssemblySequence=False,hasScorePropagation="ScorePropagation=Best",hasScoreType="ScoreType=Score",preprocessedData=[],optionList = [], DEBUG=True):
        self.DEBUG = DEBUG
        if JSONDescriptor is None:
            self.Name = Name
            self.Description = Description
            self.RuleType = RuleType
            self.isAppliedToProduct = isAppliedToProduct
            self.isAppliedToProductPart = isAppliedToProductPart
            self.isAppliedToAssemblySequence = isAppliedToAssemblySequence
            self.hasScorePropagation = str(hasScorePropagation)
            self.hasScoreType=str(hasScoreType)
            self.optionList = optionList
            self.property = None
            self.preprocessedData = preprocessedData
        else:
            # TODO: import json descriptor to setup object
            self.json2object(JSONDescriptor)

    def json2object(self,jsonDescriptor):
        """
             Function to generate a DFA object from a JSON file

             :param string jsonDescriptor: absolute path to the json file for the DFA rule

        """
        #--interpret JSON file--
        with open(jsonDescriptor, "r") as read_file:
            jsonObject = json.load(read_file)
        #--setup object--
        self.Name = jsonObject['DFARule']['Name']
        self.Description = jsonObject['DFARule']['Description']
        self.RuleType = jsonObject['DFARule']['RuleType']
        if 'True' in jsonObject['DFARule']['isAppliedToProduct']:self.isAppliedToProduct = True
        else:self.isAppliedToProduct = False
        if 'True' in jsonObject['DFARule']['isAppliedToProductPart']:self.isAppliedToProductPart = True
        else:self.isAppliedToProductPart = False
        if 'True' in jsonObject['DFARule']['isAppliedToAssemblySequence']:self.isAppliedToAssemblySequence = True
        else:self.isAppliedToAssemblySequence = False
        if '1' in jsonObject['DFARule']['hasScorePropagation']:self.hasScorePropagation = 1
        else:self.hasScorePropagation = 0
        if '1' in jsonObject['DFARule']['hasScoreType']:self.hasScoreType = 1
        else:self.hasScoreType = 0
        self.optionList = []
        for option in jsonObject['DFARule']['optionList']:
            temp = []
            temp.append(option['Name'])
            temp.append(option['Value1'])
            temp.append(option['Value2'])
            self.optionList.append(temp)
        for property in jsonObject['DFARule']['property']:
            temp = []
            temp.append(property['Name'])
            temp.append(property['Value1'])
            temp.append(property['Value2'])
            self.property=temp

        #self.preprocessedData = preprocessedData       #TODO!


    def object2json(self):
        """
            Function to generate a json file from the DFA rule
        """
        return json.dumps(self.__dict__,indent=4)

class PerformanceModel():
    """
       PerformanceModel: Class representing the complete performance model

        :param string Name: Name of the performance model
        :param string Description: Description of the performance model
        :param list OptimizationMethod: OptimizationMethod of the performance model
        :param objectlist StopConditionList: List of stop conditions
        :param list MethodPerformance: List of stop performance metrics
        :param list RAWResults: List of RAW optimization results
        :param list interpretedResults: List of interpreted optimization results
        :param objectlist parameterList: list of linked parameters
        :param objectlist ObjectiveList: list of optimization objectives
        :param objectlist ConstraintList: list of optimization constraints
    """
    def __init__(self, JSONDescriptor=None, Name="", Description="",OptimizationMethod=[],MethodPerformance=[],RAWResults=[],parameterlist = [],StopConditions = [],ObjectiveList=[],ConstraintList=[],interpretedResults=[],DEBUG=True):
        self.DEBUG = DEBUG
        if JSONDescriptor is None:
            self.Name = Name
            self.Description = Description
            #OptimizationMethod + Stop
            self.OptimizationMethod = OptimizationMethod
            self.StopConditionList = StopConditions
            # OptimizationResult
            self.MethodPerformance = MethodPerformance
            self.RAWResults = RAWResults
            self.interpretedResults=interpretedResults
            #parameter list
            self.parameterList = parameterlist
            # DESIGN TARGETS
            self.ObjectiveList = ObjectiveList
            self.ConstraintList = ConstraintList

        else:
            # TODO: import json descriptor to setup object
            self.json2object(JSONDescriptor)

    def json2object(self, jsonDescriptor):
        """
             Function to generate a performance model from a JSON file

             :param string jsonDescriptor: absolute path to the json file for performance model

        """
        # --interpret JSON file--
        with open(jsonDescriptor, "r") as read_file:
            jsonObject = json.load(read_file)
        # --setup object--
        self.Name = jsonObject['Name']
        self.Description = jsonObject['Description']
        self.OptimizationMethod = [jsonObject['OptimizationMethod'][0],jsonObject['OptimizationMethod'][1],jsonObject['OptimizationMethod'][2]]
        #stopConditions
        self.StopConditionList = []
        for stp in jsonObject["StopConditionList"]:
            stopC = StopCondition(Name=stp["Name"],Description=stp["Description"],Value=stp["Value"],Stopcriteria=stp["StopCriteria"])
            self.StopConditionList.append(stopC)

        #Method results TODO: this is not complete yet
        self.MethodPerformance = jsonObject['MethodPerformance']
        self.RAWResults = jsonObject['RAWResults']
        self.interpretedResults = jsonObject['interpretedResults']

        # DecisionVariables
        self.parameterList = []
        for var in jsonObject["parameterList"]:
            pList = []
            for p in var["parameterList"]:
                par = Parameter(Name=p["Name"], Description=p["Description"],Key=p["Key"],GUID=p["GUID"], Value=p["Value"], Stopcriteria=p["StopCriteria"])
                pList.append(par)
            decisionvariable = DecisionVariable(Name=var["Name"], Description=var["Description"], InitialValue=var["InitialValue"],MinValue=var["MinValue"], MaxValue=var["MaxValue"], Optimum=var["Optimum"],Resolution = var["Resolution"],parameters=pList)
            self.parameterList.append(decisionvariable)

        # Objectives
        self.ObjectiveList = []
        for objective in jsonObject["ObjectiveList"]:
            obj_ = Objective(Name=objective["Name"], Description=objective["Description"], ObjectiveOption=objective["ObjectiveOption"])
            self.ObjectiveList.append(obj_)

        # Constraints
        self.ConstraintList = []
        for cnstr in jsonObject["ConstraintList"]:
            cnstr_ = Constraint(Name=cnstr["Name"], Description=cnstr["Description"],Expression=cnstr["Expression"])
            self.ConstraintList.append(cnstr_)

    def clean(self):
        self.parameterList = []
        # DESIGN TARGETS
        self.ObjectiveList = []
        self.ConstraintList = []

    def object2json(self):
        """
               Function to generate a json file from the performance model
        """
        return json.dumps(self, default=lambda o: o.__dict__,indent=4)

class StopCondition():
    """
        StopCondition: Class representing the algorithm stop conditions

         :param string Name: Name of the object
         :param string Description: Description of the object
         :param string Value: Value of the object
         :param enum StopCriteria: StopCriteria of the object

    """
    def __init__(self, JSONDescriptor=None, Name="", Description="", Value=0.0, Stopcriteria="maxDepth=0", DEBUG=True):
        self.DEBUG = DEBUG
        if JSONDescriptor is None:
            self.Name = Name
            self.Description = Description
            self.Value = Value
            self.StopCriteria = Stopcriteria
        else:
            # TODO: import json descriptor to setup object
            x = 10

class Parameter():
    """
            Parameter: Class representing the parameter

             :param string Name: Name of the parameter
             :param string Description: Description of the parameter
             :param string Key: Key of the parameter
             :param string GUID: Global unique identifier of the parameter
             :param float Value: Value of the parameter
             :param enum StopCriteria: StopCriteria of the parameter
    """
    def __init__(self, JSONDescriptor=None, Name="", Description="",Key="",GUID="", Value=0.0, Stopcriteria="maxDepth=0", DEBUG=True):
        self.DEBUG = DEBUG
        if JSONDescriptor is None:
            self.Name = Name
            self.Description = Description
            self.Key = Key
            self.GUID = GUID
            self.Value = Value
            self.StopCriteria = Stopcriteria
        else:
            # TODO: import json descriptor to setup object
            x = 1

class Objective():
    """
        Objective: Class representing the optimization Objective

         :param string Name: Name of the Objective
         :param string Description: Description of the Objective
         :param enum ObjectiveOption: ObjectiveOption of the Objective
    """
    def __init__(self, JSONDescriptor=None, Name="", Description="",ObjectiveOption="Minimize", DEBUG=True):
        self.DEBUG = DEBUG
        if JSONDescriptor is None:
            self.Name = Name
            self.Description = Description
            self.ObjectiveOption = ObjectiveOption

        else:
            # TODO: import json descriptor to setup object
            x = 10

class Constraint():
    """
        Constraint: Class representing the optimization Constraint

         :param string Name: Name of the constraint
         :param string Description: Description of the constraint
         :param list Expression: Expression of the constraint
    """
    def __init__(self, JSONDescriptor=None, Name="", Description="",Expression=[], DEBUG=True):
        self.DEBUG = DEBUG
        if JSONDescriptor is None:
            self.Name = Name
            self.Description = Description
            self.Expression = Expression

        else:
            # TODO: import json descriptor to setup object
            x = 10

class DecisionVariable():
    """
        DecisionVariable: Class representing the optimization Objective

         :param string Name: Name of the DecisionVariable
         :param string Description: Description of the DecisionVariable
         :param float InitialValue: InitialValue of the DecisionVariable
         :param float MinValue: MinValue of the DecisionVariable
         :param float MaxValue: MaxValue of the DecisionVariable
         :param float Optimum: Optimum of the DecisionVariable
         :param float Resolution: Resolution of the DecisionVariable
         :param objectlist parameterList: list of linked parameters
    """
    def __init__(self, JSONDescriptor=None, Name="", Description="",InitialValue=0.0,MinValue=0.0, MaxValue=0.0, Optimum=0.0,Resolution = 0.0,parameters=[], DEBUG=True):
        self.DEBUG = DEBUG
        if JSONDescriptor is None:
            self.Name = Name
            self.Description = Description
            self.InitialValue = InitialValue
            self.MinValue = MinValue
            self.MaxValue = MaxValue
            self.Optimum = Optimum
            self.Resolution = Resolution
            self.parameterList = parameters
        else:
            # TODO: import json descriptor to setup object
            x = 1

class Product():
    """
        Product: Class representing the Product

         :param string Name: Name of the Product
         :param string STEPFile: Absolute path to the Product STEP file
         :param objectlist parameterList: list of linked parameters
    """
    def __init__(self, JSONDescriptor=None, Name="", Description="",STEPFile='',parameters = [], DEBUG=True):
        self.DEBUG = DEBUG
        if JSONDescriptor is None:
            self.Name = Name
            self.Description = Description
            self.STEPFile = STEPFile
            self.parameterList = parameters

        else:
            self.json2object(JSONDescriptor)
            x = 10

    def json2object(self,jsonDescriptor):
        """
             Function to generate a Product object from a JSON file

             :param string jsonDescriptor: absolute path to the json file for the Product

        """
        #--interpret JSON file--
        with open(jsonDescriptor, "r") as read_file:
            jsonObject = json.load(read_file)
        #--setup object--
        self.Name = jsonObject['Name']
        self.Description = jsonObject['Description']
        self.STEPFile = jsonObject['STEPFile']

        self.parameterList = []
        for par in jsonObject['parameterList']:
            p = Parameter(None, Name=par['Name'], Description=par['Description'],Key=par['Key'], GUID=par['GUID'],Value=par['Value'],Stopcriteria=par['StopCriteria'])
            self.parameterList.append(p)
        #TODO: discuss if we need to add the rules at once?

    def object2json(self):
        """
               Function to generate a json file from the Product model
        """
        return json.dumps(self, default=lambda o: o.__dict__,indent=4)

class ProductPart():
    """
        Product: Class representing the Product Part

         :param string Name: Name of the Product
         :param string ProductType: Product part type
         :param int Quantity: Number of same product parts in product
         :param objectlist parameterList: list of linked parameters
         :param objectlist partList: list of linked product parts
         :param list material: list of the material properties
    """
    def __init__(self, JSONDescriptor=None, Name="", Description="",ProductType='',Quantity=0,parameters = [],partList=[],material = [], DEBUG=True):
        self.DEBUG = DEBUG
        if JSONDescriptor is None:
            self.Name = Name
            self.Description = Description
            self.ProductType = ProductType
            self.Quantity = Quantity
            self.parameterList = parameters
            self.partList = partList
            self.material = material    #TODO: Stiffness not yet correct

        else:
            self.json2object(JSONDescriptor)
            x = 10

    def json2object(self,jsonDescriptor):
        """
             Function to generate a ProductPart object from a JSON file

             :param string jsonDescriptor: absolute path to the json file for the ProductPart

        """

        #--interpret JSON file--
        with open(jsonDescriptor, "r") as read_file:
            jsonObject = json.load(read_file)
        #--setup object--
        self.Name = jsonObject['Name']
        self.Description = jsonObject['Description']
        self.ProductType = jsonObject['ProductType']
        self.Quantity = jsonObject['Quantity']
        self.parameterList = []
        for par in jsonObject['parameterList']:
            p = Parameter(None, Name=par['Name'], Description=par['Description'], Key=par['Key'], GUID=par['GUID'],Value=par['Value'], Stopcriteria=par['StopCriteria'])
            self.parameterList.append(p)
        self.partList = []
        for subPart in jsonObject['partList']:
            p = self.getInstance(subPart)
            self.partList.append(p)
        self.material = jsonObject['material']  # TODO: Stiffness not yet correct

    def object2json(self):
        """
               Function to generate a json file from the ProductPart model
        """
        return json.dumps(self, default=lambda o: o.__dict__,indent=4)

    def getInstance(self,subPart):
        """
               Function to fetch a ProductPart model recursively
        """
        part_temp = ProductPart(Name=subPart['Name'],Description=subPart['Description'],ProductType=subPart['ProductType'],Quantity=subPart['Quantity'])

        part_temp.parameterList = []
        for par in subPart['parameterList']:
            p = Parameter(None, Name=par['Name'], Description=par['Description'], Key=par['Key'], GUID=par['GUID'], Value=par['Value'], Stopcriteria=par['StopCriteria'])
            part_temp.parameterList.append(p)
        part_temp.partList = []
        for subPart in subPart['partList']:
            p = self.getInstance(subPart)
            part_temp.partList.append(p)
        part_temp.material = subPart['material']  # TODO: Stiffness not yet correct
        return part_temp

class AssemblySequence():
    """
        Product: Class representing the AssemblySequence

         :param string Name: Name of the AssemblySequence
         :param string Description: Description of the AssemblySequence
         :param list AssemblyMetric: AssemblyMetric identification
         :param objectList assemblyOptions: list of assemblyOption objects
    """

    def __init__(self, JSONDescriptor=None, Name="", Description="", AssemblyMetric=[],AssemblyOptions=None, DEBUG=True):
        self.DEBUG = DEBUG
        if JSONDescriptor is None:
            self.Name = Name
            self.Description = Description
            self.AssemblyMetric = AssemblyMetric
            self.AssemblyOptions = AssemblyOptions

        else:
            self.rootAssembly=None
            self.childActive = False
            self.json2object(JSONDescriptor)
            x = 10

    def setAssemblyOption(self, Option):
        try:
            # operations
            operationList = []
            for operation in Option['Operations']:
                op = Operation(Name=operation['Name'], Description="NOT IN MM")
                op.OperationMetric = [operation['OperationMetric'][0], operation['OperationMetric'][1]]  # TODO: in MM, multiple metrics can be added -> is this necessary?
                # operators
                operatorList = []
                for operator in operation['Operators']:
                    operator_temp = Operator(Name=operator['Name'],CostPerHour=operator['CostPerHour'],Height=operator['Height'],MaxLiftingWeight=operator['MaxLiftingWeight'],Reach=operator['Reach'])
                    operatorList.append(operator_temp)
                op.Operators = operatorList
                # Tools
                ToolList = []
                for tool in operation['Tool']:
                    requiredOperators = []
                    for r_operator in operation['Operators']:
                        requiredOperators.append(r_operator['Name'])  # TODO: only name reference added, ok?
                    Tool_temp = Tool(Name=tool['Name'],Mass=tool['Mass'],Cost=tool['Cost'],Orientation=tool['Orientation'],Geometry=tool['Geometry'],RequiredOperators=requiredOperators)
                    #Tool_temp = [tool.hasName, tool.hasMass, tool.hasCost, tool.hasOrientation.hasData,tool.hasGeometry.hasDataInStepFile, requiredOperators]
                    ToolList.append(Tool_temp)
                op.Tool = ToolList
                # Fastners
                FastnersList = []
                for fastner in operation['Fastners']:
                    FastnersList.append(fastner[0])  # TODO: only name reference added, ok?
                op.Fastners = FastnersList
                operationList.append(op)

            # -- define the assemblyOption --
            ASO = AssemblyOption(Name=Option['Name'],Operations=operationList)

            #Linking the recursive operations!
            subAssemblyOptions = []
            for recursive_option in Option['RecursiveAssemblyOption']:
                subAssemblyOptions.append(self.setAssemblyOption(recursive_option))

            ASO.RecursiveAssemblyOption = subAssemblyOptions

            return ASO

        except:
            return []

    def json2object(self, jsonDescriptor):
        """
             Function to generate a AssemblySequence object from a JSON file

             :param string jsonDescriptor: absolute path to the json file for the AssemblySequence

        """

        # --interpret JSON file--
        with open(jsonDescriptor, "r") as read_file:
            jsonObject = json.load(read_file)
            self.Name=jsonObject['Name']
            self.Description = jsonObject['Description']
            self.AssemblyMetric = [jsonObject['AssemblyMetric'][0],jsonObject['AssemblyMetric'][1]]

            # -- recursive AssemblyOptions --
            self.AssemblyOptions = self.setAssemblyOption(Option=jsonObject['AssemblyOptions'])




    def object2json(self):
        """
               Function to generate a json file from the AssemblySequence model
        """
        return json.dumps(self, default=lambda o: o.__dict__, indent=4)

class AssemblyOption():
    """
            AssemblyOption: Class representing the AssemblyOption

             :param string Name: Name of the AssemblyOption
             :param float Description: Cost of the AssemblyOption
             :param ObjectList Operations: OperationList of the AssemblyOption
             :param list OperatorMetric: OperatorMetric of the AssemblyOption
             :param ObjectList Operators: List of required operators of the AssemblyOption
             :param objectList ToolList: Required tools of the AssemblyOption
             :param objectList RecursiveAssemblyOption: Nested AssemblyOption

    """
    def __init__(self, JSONDescriptor=None, Name="", Description="",Operations=[],RecursiveAssemblyOption=[], DEBUG=True):
        self.DEBUG = DEBUG
        if JSONDescriptor is None:
            self.Name = Name
            self.Description = Description
            self.Operations = Operations
            self.RecursiveAssemblyOption = RecursiveAssemblyOption

        else:
            # TODO: import json descriptor to setup object
            x = 1

    def object2json(self):
        """
               Function to generate a json file from the AssemblySequence model
        """
        return json.dumps(self, default=lambda o: o.__dict__, indent=4)

class Operator():
    """
            Operator: Class representing the Operator

             :param string Name: Name of the Operator
             :param float CostPerHour: Cost of the Operator
             :param float Height: Height of the Operator in cm
             :param float MaxLiftingWeight: Maximum lifting capabilities of the Operator
             :param float Reach: Reach capabilities of the Operator

    """
    def __init__(self, JSONDescriptor=None, Name="", CostPerHour=70.0,Height=180.0,MaxLiftingWeight=50, Reach=155.0, DEBUG=True):
        self.DEBUG = DEBUG
        if JSONDescriptor is None:
            self.Name = Name
            self.CostPerHour = CostPerHour
            self.Height = Height
            self.MaxLiftingWeight = MaxLiftingWeight
            self.Reach = Reach

        else:
            # TODO: import json descriptor to setup object
            x = 1

    def object2json(self):
        """
               Function to generate a json file from the AssemblySequence model
        """
        return json.dumps(self, default=lambda o: o.__dict__, indent=4)

class Tool():
    """
            Tool: Class representing the Tool   tool.hasName, tool.hasMass, tool.hasCost, tool.hasOrientation.hasData,
                                 tool.hasGeometry.hasDataInStepFile, requiredOperators

             :param string Name: Name of the Tool
             :param float Mass: Mass of the Tool
             :param float Cost: Cost of the Tool
             :param string Orientation: Orientation data of the Tool
             :param string hasGeometry: hasGeometry info of the Tool
             :param List RequiredOperators: List of required tool operators (referenced by name)

    """
    def __init__(self, JSONDescriptor=None, Name="", Mass=70.0,Cost=500.0,Orientation="",Geometry="", RequiredOperators=[], DEBUG=True):
        self.DEBUG = DEBUG
        if JSONDescriptor is None:
            self.Name = Name
            self.Mass = Mass
            self.Cost = Cost
            self.Orientation = Orientation
            self.Geometry = Geometry
            self.RequiredOperators = RequiredOperators

        else:
            # TODO: import json descriptor to setup object
            x = 1

    def object2json(self):
        """
               Function to generate a json file from the AssemblySequence model
        """
        return json.dumps(self, default=lambda o: o.__dict__, indent=4)

class Operation():       #
    """
        Product: Class representing the Operation

         :param string Name: Name of the Operation
         :param string Description: Description of the Operation
         :param list Fastner: Fastner identification
         :param list OperationMetric: Operation Metric identification
         :param list Operators: Operators identification
         :param list Tool: Tool identification
         :param objectList AssemblyOperationList: list of assemblyOperations
    """

    def __init__(self, JSONDescriptor=None, Name="", Description="", Fastners=[],OperationMetric=[],Operators = [], Tool=[], DEBUG=True):
        self.DEBUG = DEBUG
        if JSONDescriptor is None:
            self.Name = Name
            self.Description = Description
            self.Fastners = Fastners
            self.OperationMetric = OperationMetric
            self.Operators = Operators
            self.Tool = Tool


        else:
            self.json2object(JSONDescriptor)
            x = 10

    def json2object(self, jsonDescriptor):
        """
             Function to generate a Operation object from a JSON file

             :param string jsonDescriptor: absolute path to the json file for the Operation

        """

        # --interpret JSON file--
        with open(jsonDescriptor, "r") as read_file:
            jsonObject = json.load(read_file)


    def object2json(self):
        """
               Function to generate a json file from the Operation model
        """
        return json.dumps(self, default=lambda o: o.__dict__, indent=4)


#---------------------Ontology & contracts interface objects--------------------------

class Contract():       #
    """
        Contract: Class representing the Contract

         :param string Name: Name of the Contract
         :param string Description: Description of the Contract
         :param list Assmuptions: list of contract assumptions
         :param list Guarantees: list of contract guarantees
    """

    def __init__(self, JSONDescriptor=None, Name="", Description="", Assumptions=[],Guarantees=[], DEBUG=True):
        self.DEBUG = DEBUG
        if JSONDescriptor is None:
            self.Name = Name
            self.Description = Description
            self.Assumptions = Assumptions
            self.Guarantees = Guarantees


        else:
            self.json2object(JSONDescriptor)
            x = 10

    def json2object(self, jsonDescriptor):
        """
             Function to generate a Contract object from a JSON file

             :param string jsonDescriptor: absolute path to the json file for the Contract

        """

        # --interpret JSON file--
        with open(jsonDescriptor, "r") as read_file:
            jsonObject = json.load(read_file)

            self.Name = jsonObject['Name']
            self.Description = jsonObject['Description']
            self.Assumptions = []
            for asmpt in jsonObject['Assumptions']:
                temp = []
                dv = DomainVariable(Name=asmpt[0]["Name"],Description=asmpt[0]["Description"],Unit=asmpt[0]["Unit"])
                temp.append(dv)
                temp.append(asmpt[1])
                temp.append(asmpt[2])
                self.Assumptions.append(temp)

            self.Guarantees = []
            for grnt in jsonObject['Guarantees']:
                temp = []
                dv = DomainVariable(Name=grnt[0]["Name"], Description=grnt[0]["Description"], Unit=grnt[0]["Unit"])
                temp.append(dv)
                temp.append(grnt[1])
                temp.append(grnt[2])
                self.Guarantees.append(temp)


    def object2json(self):
        """
               Function to generate a json file from the Contract model
        """
        return json.dumps(self, default=lambda o: o.__dict__, indent=4)

class Ontology():       #
    """
        Ontology: Class representing the Ontology

         :param string Name: Name of the Ontology
         :param string Description: Description of the Ontology
         :param objectlist Relations: list of relation objects
    """

    def __init__(self, JSONDescriptor=None, Name="", Description="", Relations=[], DEBUG=True):
        self.DEBUG = DEBUG
        if JSONDescriptor is None:
            self.Name = Name
            self.Description = Description
            self.Relations = Relations


        else:
            self.json2object(JSONDescriptor)
            x = 10

    def json2object(self, jsonDescriptor):
        """
             Function to generate a Ontology object from a JSON file

             :param string jsonDescriptor: absolute path to the json file for the Ontology

        """

        # --interpret JSON file--
        with open(jsonDescriptor, "r") as read_file:
            jsonObject = json.load(read_file)

            self.Name = jsonObject['Name']
            self.Description = jsonObject['Description']
            self.Relations = []
            for relation in jsonObject['Relations']:
                temp = []
                source = DomainVariable(Name=relation['Source']['Name'],
                                        Description=relation['Source']['Description'],
                                        Unit=relation['Source']['Unit'])
                destination = DomainVariable(Name=relation['Destination']['Name'],
                                        Description=relation['Destination']['Description'],
                                        Unit=relation['Destination']['Unit'])
                r = Relation(Name=relation['Name'], Description=relation['Description'], Source=source,Destination=destination,SensitivityDirection=relation['SensitivityDirection'],Weight=relation['Weight'])
                self.Relations.append(r)


    def object2json(self):
        """
               Function to generate a json file from the Ontology model
        """
        return json.dumps(self, default=lambda o: o.__dict__, indent=4)

class Relation():       #
    """
        Relation: Class representing the Relation

         :param string Name: Name of the Relation
         :param string Description: Description of the Relation
         :param object Source: Source object of the relation.
         :param object Destination: Destination object of the relation
         :param string SensitivityDirection: SensitivityDirection of the relation (None = No directional sensitivity information)
         :param float Weight: Sensitivity weight of the relation (None = No sensitivity weight information)
    """

    def __init__(self, JSONDescriptor=None, Name="", Description="", Source = None,Destination = None,SensitivityDirection = None,Weight = 0.0, DEBUG=True):
        self.DEBUG = DEBUG
        if JSONDescriptor is None:
            self.Name = Name
            self.Description = Description
            self.Source = Source
            self.Destination = Destination
            self.SensitivityDirection = SensitivityDirection
            self.Weight = Weight


        else:
            self.json2object(JSONDescriptor)
            x = 10

    def json2object(self, jsonDescriptor):
        """
             Function to generate a Relation object from a JSON file

             :param string jsonDescriptor: absolute path to the json file for the Relation

        """

        # --interpret JSON file--
        with open(jsonDescriptor, "r") as read_file:
            jsonObject = json.load(read_file)


    def object2json(self):
        """
               Function to generate a json file from the Relation model
        """
        return json.dumps(self, default=lambda o: o.__dict__, indent=4)

class DomainVariable():       #
    """
        DomainVariable: Class representing the DomainVariable

         :param string Name: Name of the DomainVariable
         :param string Description: Description of the DomainVariable
         :param string Unit: Unit of the DomainVariable
    """

    def __init__(self, JSONDescriptor=None, Name="", Description="", Unit="", DEBUG=True):
        self.DEBUG = DEBUG
        if JSONDescriptor is None:
            self.Name = Name
            self.Description = Description
            self.Unit = Unit


        else:
            self.json2object(JSONDescriptor)
            x = 10

    def json2object(self, jsonDescriptor):
        """
             Function to generate a DomainVariable object from a JSON file

             :param string jsonDescriptor: absolute path to the json file for the DomainVariable

        """

        # --interpret JSON file--
        with open(jsonDescriptor, "r") as read_file:
            jsonObject = json.load(read_file)


    def object2json(self):
        """
               Function to generate a json file from the DomainVariable model
        """
        return json.dumps(self, default=lambda o: o.__dict__, indent=4)


#interface objects
from .InterfaceObjects import ASG,StopCondition
#import helper functions
from .HelperFunctions import *

#-----------------------------------------------------------------------------------------------------------
#
#   ASG extension
#
#       TODO:
#               - Discuss: Do we need to provide the rules in the getASG()
#               - Discuss: Do we need to provide the rules in the setASG()
#               - Discuss: Can we assume that the assembly system is in place when performing setASG()
#               - Discuss: Do we need seperate functions to UPDATE and ADD DFA rules?
#
# -----------------------------------------------------------------------------------------------------------

def getASG(AssemblySystemName,model,KBPath):
    """
          Function to fetch the Assembly Sequence Generation (ASG) configuration using the name as identifier.
          The function returns an interface block of the ASG

          :param string AssemblySystemName: Name identifier of the assemlbySystem
          :param object model: Metamodel instance model
          :param string KBPath: Absolute path to the metamodel instance model

          :return object ASGModel: Interface object of the ASG (-1 = error)
    """

    found = 0
    InterfaceObject = ASG(None) #interface object placeholder
    if len(model.includesAssemblySystem.items) != 0:
        #match the assemblySystem by Name
        for assemblySystem in model.includesAssemblySystem.items:
            if AssemblySystemName == assemblySystem.hasName:
                found=1
                #fetch ASG and setup ASG return
                ASG_temp = assemblySystem.hasAssemblyAlgorithmConfiguration
                InterfaceObject.Name = ASG_temp.hasName
                InterfaceObject.Description = ASG_temp.hasDescription
                InterfaceObject.ProcessingType = ASG_temp.hasProcessingType.name

                # -- ASG GENERATOR --
                try:
                        InterfaceObject.Generator=[ASG_temp.hasSinglePartGenerator.hasName, 'SinglePartGenerator']
                except:
                    x = "No SinglePartGenerator available"
                try:
                        InterfaceObject.Generator=[ASG_temp.hasMultiPartGenerator.hasName, 'MultiPartGenerator']
                except:
                    x = "No MultiPartGenerator available"

                # try-catch structure to fetch the ASG SELECTOR
                try:
                        InterfaceObject.Selector=[ASG_temp.hasDefaultSelector.hasName, 'DefaultSelector']
                except:
                    x = "No Default selector available"
                try:
                        InterfaceObject.Selector=[ASG_temp.hasRuleSelector.hasName, 'RuleSelector']             #TODO: discuss if we need to fetch the rules
                except:
                    x = "No Rule selector available"

                # try-catch structure to fetch the ASG EVALUATOR
                try:
                        InterfaceObject.Evaluator=[ASG_temp.hasUniformEvaluator.hasName,'UniformEvaluator']
                except:
                    x = "No uniform evalutor available"
                try:
                        InterfaceObject.Evaluator=[ASG_temp.hasRuleEvaluator.hasName, 'RuleEvaluator']
                except:
                    x = "No Rule selector available"

                # -- Terminator --
                terms = []
                try:
                    for terminator in ASG_temp.hasTerminator.hasStopCriterion.items:
                        # -- stopconditions --
                        s = StopCondition(None, terminator.hasName, terminator.hasDescription,terminator.hasValue, terminator.hasStopCriterion.name)

                        #add to criteria
                        terms.append(s)
                    InterfaceObject.StopConditionList = terms
                except:
                    x = "No Stop criterion available"

            else:
                x = 1

        if found:
            return InterfaceObject
        else:
            return -2
    else:
        return -2  # no assembly systems!

def updateASG(AssemblySystemName,interfaceObject,model,KBPath):
    """
          Function to update the  complete Assembly Sequence Generation (ASG) configuration using the name as identifier.
          The ASG interface object is used to interface with the function.
          The function returns whether or not the function is performed correctly

          :param string AssemblySystemName: Name identifier of the assemlbySystem
          :param object ASGModel: Interface object of the ASG
          :param object model: Metamodel instance model
          :param string KBPath: Absolute path to the metamodel instance model

          :return int Error: -1 = error, 1= function performed correcly
    """

    found = 0
    if len(model.includesAssemblySystem.items) != 0:
        #match the assemblySystem by Name
        for assemblySystem in model.includesAssemblySystem.items:
            if AssemblySystemName == assemblySystem.hasName:
                found=1
                # -- ASG setup --
                assemblySystem.hasAssemblyAlgorithmConfiguration.hasName = interfaceObject.Name
                assemblySystem.hasAssemblyAlgorithmConfiguration.hasDescription = interfaceObject.Description
                assemblySystem.hasAssemblyAlgorithmConfiguration.hasProcessingType = interfaceObject.ProcessingType

                # -- ASG GENERATOR --
                try:
                    assemblySystem.hasAssemblyAlgorithmConfigurationhasSinglePartGenerator.hasSinglePartGenerator.hasName = interfaceObject.Generator[0]
                except:
                    x = "No SinglePartGenerator available"
                try:
                    assemblySystem.hasAssemblyAlgorithmConfigurationhasSinglePartGenerator.hasMultiPartGenerator.hasName = interfaceObject.Generator[0]
                except:
                    x = "No MultiPartGenerator available"

                # -- ASG SELECTOR --
                try:
                    assemblySystem.hasAssemblyAlgorithmConfigurationhasSinglePartGenerator.hasDefaultSelector.hasName =  interfaceObject.Selector[0]
                except:
                    x = "No Default selector available"
                try:
                    assemblySystem.hasAssemblyAlgorithmConfigurationhasSinglePartGenerator.hasRuleSelector.hasName = interfaceObject.Selector[0]
                    #TODO: discuss if we need to update the rules or via dedicated function!
                except:
                    x = "No Rule selector available"

                # -- ASG EVALUATOR --
                try:
                    assemblySystem.hasAssemblyAlgorithmConfigurationhasSinglePartGenerator.hasUniformEvaluator.hasName = interfaceObject.Evaluator[0]
                except:
                    x = "No uniform evalutor available"
                try:
                    assemblySystem.hasAssemblyAlgorithmConfigurationhasSinglePartGenerator.hasRuleEvaluator.hasName = interfaceObject.Evaluator[0]
                    # TODO: discuss if we need to update the rules or via dedicated function!
                except:
                    x = "No Rule selector available"



                # -- Terminator container --
                for terminator in assemblySystem.hasAssemblyAlgorithmConfiguration.hasTerminator.hasStopCriterion.items:
                    for stp in interfaceObject.StopConditionList:
                        if stp.Name==terminator.hasName:
                            terminator.hasName=stp.Name
                            terminator.hasDescription=stp.Description
                            terminator.hasValue=stp.Value
                            terminator.hasStopCriterion=stp.StopCriteria

            else:
                return -1

        if found:
            updateKBv6(KBPath, model)
            return 1
        else:
            return -1
    else:
        return -1  # no Optimization Problem!

def setASG( AssemblySystemName, InterfaceObject,model,KBPath,path_ecore,MM,API):
    """
          Function to set a new Assembly Sequence Generation (ASG) configuration to a new AssemblySystem
          The performance interface object is used to interface with the function.
          The function returns whether or not the function is performed correctly

          :param object ASGModel: Interface object of the ASG model
          :param object model: Metamodel instance model
          :param string KBPath: Absolute path to the metamodel instance model
          :param string KBPath: Absolute path to the metamodel ecore model

          :return int Error: -1 = error, 1= function performed correcly
    """

    # -- create new AssemblySystem --
    AS = API.create_connected("AssemblySystem")
    AS.hasName = AssemblySystemName
    # -- create new ASG --
    ASG_ = API.create("AssemblyAlgorithmConfiguration")
    ASG_.hasName = InterfaceObject.Name
    ASG_.hasDescription = InterfaceObject.Description
    #ASG.hasProcessingType = InterfaceObject.ProcessingType                 #TODO: check how to set this

    if 'SinglePartGenerator' in InterfaceObject.Generator[1]:
        # -- ASG GENERATOR --
        try:
            ASG_.hasSinglePartGenerator.hasName = InterfaceObject.Generator[0]
        except:
            x = "No SinglePartGenerator available"
            # -- create new class if not existing --
            g_ = API.create("SinglePartGenerator")
            g_.hasName = InterfaceObject.Generator[0]
            ASG_.hasSinglePartGenerator = g_

    if 'MultiPartGenerator' in InterfaceObject.Generator[1]:
        # -- ASG GENERATOR --
        try:
            ASG_.hasMultiPartGenerator.hasName = InterfaceObject.Generator[0]
        except:
            x = "No MultiPartGenerator available"
            # -- create new class if not existing --
            g_ = API.create("MultiPartGenerator")
            g_.hasName = InterfaceObject.Generator[0]
            ASG_.hasMultiPartGenerator = g_

    if 'DefaultSelector' in InterfaceObject.Selector[1]:
        # -- ASG SELECTOR --
        try:
            ASG_.hasDefaultSelector.hasName = InterfaceObject.Selector[0]
        except:
            x = "No Default selector available"
            # -- create new class if not existing --
            g_ = API.create("DefaultSelector")
            g_.hasName = InterfaceObject.Selector[0]
            ASG_.hasDefaultSelector = g_

    if 'RuleSelector' in InterfaceObject.Selector[1]:
        # -- ASG SELECTOR --
        try:
            ASG_.hasRuleSelector.hasName = InterfaceObject.Selector[0]
        except:
            x = "No RuleSelector available"
            # -- create new class if not existing --
            g_ = API.create("RuleSelector")
            g_.hasName = InterfaceObject.Selector[0]
            ASG_.hasRuleSelector = g_

    if 'UniformEvaluator' in InterfaceObject.Evaluator[1]:
        # -- ASG EVALUATOR --
        try:
            ASG_.hasUniformEvaluator.hasName = InterfaceObject.Evaluator[0]
        except:
            x = "No uniform evaluator available"
            # -- create new class if not existing --
            g_ = API.create("UniformEvaluator")
            g_.hasName = InterfaceObject.Evaluator[0]
            ASG_.hasDefaultSelector = g_

    if 'RuleEvaluator' in InterfaceObject.Evaluator[1]:
        # -- ASG EVALUATOR --
        try:
            ASG_.hasRuleEvaluator.hasName = InterfaceObject.Evaluator[0]
        except:
            x = "No RuleEvaluator available"
            # -- create new class if not existing --
            g_ = API.create("RuleEvaluator")
            g_.hasName = InterfaceObject.Evaluator[0]
            ASG_.hasRuleEvaluator = g_

    # -- stop conditions --
    terminator_container = API.create("Terminator")
    terminator_container.hasName = "Default"
    ASG_.hasTerminator = terminator_container
    for stp in InterfaceObject.StopConditionList:
        #terminator = API.create_custom(e_class="StopCriterion",parent=terminator_container)
        terminator= API.create_noPlatformRoot("StopCriterion")
        terminator.hasName = stp.Name
        terminator.hasDescription = stp.Description
        terminator.hasValue = stp.Value
        #terminator.hasStopCriterion = stp.StopCriteria         #TODO:assigns currently a string!
        terminator_container.hasStopCriterion.append(terminator)

    AS.hasAssemblyAlgorithmConfiguration = ASG_

    updateKBv6(KBPath,model)


#interface objects
from .InterfaceObjects import ASG,StopCondition,DFARule
#import helper functions
from .HelperFunctions import *

#-----------------------------------------------------------------------------------------------------------
#
#   DFA extension
#
#       TODO:
#
# -----------------------------------------------------------------------------------------------------------
def getASG_DFARules( AssemblySystemName, component,model):
    """
             Function to fetch the linked DFA rules within the ASG component
             The function returns a list of interface blocks of the DFA rule

             :param string AssemblySystemName: Name identifier of the assemlbySystem
             :param object model: Metamodel instance model
             :param string component: identify the ASG element ("Selector"|"Evaluator")

             :return objectList DFARuleList: list of interface object of the DFA rule (-1 = error)
    """
    # match the AssemblySystem-of-interest
    if len(model.includesAssemblySystem.items) != 0:
        # match the assemblySystem by Name
        for assemblySystem in model.includesAssemblySystem.items:
            if AssemblySystemName == assemblySystem.hasName:
                # check if ASG config exist
                if assemblySystem.hasAssemblyAlgorithmConfiguration is not None:
                    ASG_temp = assemblySystem.hasAssemblyAlgorithmConfiguration
                    if component == "Selector":
                        # check if rule selector is available
                        try:
                            cmp = ASG_temp.hasRuleSelector
                            DFARuleList = getDFArules(cmp)
                            return DFARuleList
                        except:
                            x = "No Rule selector available"
                            pass

                    if component == "Evaluator":
                        # check if rule selector is available
                        try:
                            cmp = ASG_temp.hasRuleEvaluator
                            DFARuleList = getDFArules(cmp)
                            return DFARuleList
                        except:
                            x = "No Rule evaluator available"
                            pass

                else:
                    return -1  # error - No ASG = No DFA rules

def updateASG_DFARules( AssemblySystemName, component,rule,model,KBPath):
    """
      Function to update the linked DFA rules within the ASG component
      The DFA rule interface object is used to interface with the function.
      The function returns whether or not the function is performed correctly

      :param string AssemblySystemName: Name identifier of the assemlbySystem
      :param string component: identify the ASG element ("Selector"|"Evaluator")
      :param object DFARule: Interface object of the DFA rule model
      :param object model: Metamodel instance model
      :param string KBPath: Absolute path to the metamodel instance model

      :return int Error: -1 = error, 1= function performed correcly
    """
    # match the AssemblySystem-of-interest
    DFARuleList = []
    if len(model.includesAssemblySystem.items) != 0:
        # match the assemblySystem by Name
        for assemblySystem in model.includesAssemblySystem.items:
            if AssemblySystemName == assemblySystem.hasName:
                # check if ASG config exist
                if assemblySystem.hasAssemblyAlgorithmConfiguration is not None:
                    ASG_temp = assemblySystem.hasAssemblyAlgorithmConfiguration
                    if component=="Selector":
                        #check if rule selector is available
                        try:
                            cmp = ASG_temp.hasRuleSelector
                            updateDFArule(cmp,rule)
                            updateKBv6(KBPath,model)
                        except:
                            x = "No Rule selector available"
                            pass

                    if component=="Evaluator":
                        #check if rule selector is available
                        try:
                            cmp = ASG_temp.hasRuleEvaluator
                            updateDFArule(cmp, rule)
                            updateKBv6(KBPath,model)
                        except:
                            x = "No Rule evaluator available"
                            pass

                else:
                    return -1   #error - No ASG = No DFA rules
    else:
        return -1

def getDFArules( ruleContainer):
    """
      For internal use only!
    """
    ruleList = []
    for DFArule in ruleContainer.hasDFARule.items:
        ruleDefinition = []

        x = str(type(DFArule))
        if 'DFA_MaxValue' in x:
            ruleDefinition.append("DFA_MaxValue")
        elif 'DFA_MinValue' in x:
            ruleDefinition.append("DFA_MinValue")
        elif 'DFA_BoundValue' in x:
            ruleDefinition.append("DFA_BoundValue")
        elif 'DFA_StandardParts' in x:
            ruleDefinition.append("DFA_StandardParts")
        elif 'DFA_Translation' in x:
            ruleDefinition.append(DFArule.hasTranslationType)
        elif 'DFA_Gravity' in x:
            ruleDefinition.append("DFA_Gravity")
        elif 'DFA_Misplacement' in x:
            ruleDefinition.append("DFA_Misplacement")
        elif 'DFA_Tool' in x:
            ruleDefinition.append("DFA_Tool")
        elif 'DFA_Reachability' in x:
            ruleDefinition.append("DFA_Reachability")
        elif 'DFA_Visibility' in x:
            ruleDefinition.append("DFA_Visibility")
        else:
            ruleDefinition.append("Generic_DFA")
        ruleDefinition.append(DFArule.hasName)
        ruleDefinition.append(DFArule.hasDescription)
        ruleDefinition.append(DFArule.isAppliedToProduct)
        ruleDefinition.append(DFArule.isAppliedToProductPart)
        ruleDefinition.append(DFArule.isAppliedToAssemblySequence)
        ruleDefinition.append(DFArule.hasScorePropagation)
        ruleDefinition.append(DFArule.hasScoreType)
        if DFArule.hasPreprocessedData is not None:
            ruleDefinition.append([DFArule.hasPreprocessesData.items[0].hasDataType,
                                   DFArule.hasPreprocessesData.items[0].hasDataReference])
        optionList = []
        if len(DFArule.hasOption.items) != 0:
            for option in DFArule.hasOption.items:
                optList = []
                try:
                    if option.hasValue is not None:
                        optList.append(option.hasValue)
                except:
                    x = 10
                try:
                    if option.hasX is not None:
                        optList.append(option.hasX)
                except:
                    x = 10
                try:
                    if option.hasY is not None:
                        optList.append(option.hasY)
                except:
                    x = 10
                try:
                    if option.hasZ is not None:
                        optList.append(option.hasZ)
                except:
                    x = 10
                try:
                    if option.hasAdHocFeature is not None:
                        optList.append(option.hasAdHocFeature)
                except:
                    x = 10

                optionList.append([option.hasName, optList])
        ruleDefinition.append(optionList)

        # check the subtyping
        try:
            if DFArule.hasProperty is not None:  # AS-indeptendent

                try:
                    if DFArule.hasValue is not None:
                        ruleDefinition.append([DFArule.hasProperty, DFArule.hasValue])
                except:
                    x = 10

                try:
                    if DFArule.hasMinValue is not None:
                        ruleDefinition.append([DFArule.hasProperty, DFArule.hasMinValue, DFArule.hasMaxValue])
                except:
                    x = 10

        except:  # AS-dependent
            x = 1
            # TODO: Add the AS-Dependent elements

        # transform to interface object for DFA rule
        rule_interface = DFARule(RuleType=ruleDefinition[0], Name=ruleDefinition[1],
                                 Description=ruleDefinition[2], isAppliedToProduct=ruleDefinition[3],
                                 isAppliedToProductPart=ruleDefinition[4],
                                 isAppliedToAssemblySequence=ruleDefinition[5],
                                 hasScorePropagation=ruleDefinition[6], hasScoreType=ruleDefinition[7])
        if DFArule.hasPreprocessedData is not None:
            rule_interface.preprocessedData = ruleDefinition[8]
        rule_interface.optionList = optionList
        try:
            if DFArule.hasProperty is not None:
                rule_interface.property = ruleDefinition[9]
        except:
            x = 10

        # ruleList.append(ruleDefinition)
        ruleList.append(rule_interface)

    return ruleList

def addASG_DFARule( AssemblySystemName, component,rule,model,KBPath,API):
    #     """
    #       Function to add a single DFA rules within the ASG component
    #       The DFA rule interface object is used to interface with the function.
    #       The function returns whether or not the function is performed correctly
    #
    #       :param string AssemblySystemName: Name identifier of the assemlbySystem
    #       :param string component: identify the ASG element ("Selector"|"Evaluator")
    #       :param object DFARule: Interface object of the DFA rule model
    #       :param object model: Metamodel instance model
    #       :param string KBPath: Absolute path to the metamodel instance model
    #       :param object API: API object
    #
    #       :return int Error: -1 = error, 1= function performed correcly
    #     """
    #

    # match the AssemblySystem-of-interest
    if len(model.includesAssemblySystem.items) != 0:
        # match the assemblySystem by Name
        for assemblySystem in model.includesAssemblySystem.items:
            if AssemblySystemName == assemblySystem.hasName:
                # check if ASG config exist
                if assemblySystem.hasAssemblyAlgorithmConfiguration is not None:
                    ASG_temp = assemblySystem.hasAssemblyAlgorithmConfiguration
                    if component=="Selector":
                        #check if rule selector is available
                        try:
                            cmp = ASG_temp.hasRuleSelector
                            addDFArule(cmp,rule,API)
                            updateKBv6(KBPath,model)
                        except:
                            x = "No Rule selector available"
                            pass

                    if component=="Evaluator":
                        #check if rule selector is available
                        try:
                            cmp = ASG_temp.hasRuleEvaluator
                            addDFArule(cmp, rule,API)
                            updateKBv6(KBPath,model)
                        except:
                            x = "No Rule evaluator available"
                            pass

                else:
                    return -1   #error - No ASG = No DFA rules
    else:
        return -1

def addASG_DFARules( AssemblySystemName, component,ruleList,model,KBPath,API):
    #     """
    #       Function to add multiple DFA rules within the ASG component
    #       The DFA rule interface object is used to interface with the function.
    #       The function returns whether or not the function is performed correctly
    #
    #       :param string AssemblySystemName: Name identifier of the assemlbySystem
    #       :param string component: identify the ASG element ("Selector"|"Evaluator")
    #       :param object DFARule: Interface object of the DFA rule model
    #       :param object model: Metamodel instance model
    #       :param string KBPath: Absolute path to the metamodel instance model
    #       :param object API: API object
    #
    #       :return int Error: -1 = error, 1= function performed correcly
    #     """
    #

    # match the AssemblySystem-of-interest

    if len(model.includesAssemblySystem.items) != 0:
        # match the assemblySystem by Name
        for assemblySystem in model.includesAssemblySystem.items:
            if AssemblySystemName == assemblySystem.hasName:
                # check if ASG config exist
                if assemblySystem.hasAssemblyAlgorithmConfiguration is not None:
                    ASG_temp = assemblySystem.hasAssemblyAlgorithmConfiguration
                    if component=="Selector":
                        #check if rule selector is available
                        try:
                            cmp = ASG_temp.hasRuleSelector
                            for rule in ruleList:
                                addDFArule(cmp,rule,API)
                            updateKBv6(KBPath,model)
                        except:
                            x = "No Rule selector available"
                            pass

                    if component=="Evaluator":
                        #check if rule selector is available
                        try:
                            cmp = ASG_temp.hasRuleEvaluator
                            for rule in ruleList:
                                addDFArule(cmp, rule, API)
                            updateKBv6(KBPath,model)
                        except Exception as e:
                            x = "No Rule evaluator available"
                            pass

                else:
                    return -1   #error - No ASG = No DFA rules
    else:
        return -1

def updateDFArule(self, ruleContainer, rule):
    """
      For internal use only!
    """
    for DFArule in ruleContainer.hasDFARule.items:
        if DFArule.hasName == rule[1]:
            DFArule.hasName = rule.Name
            DFArule.hasDescription = rule.Description
            DFArule.isAppliedToProduct = rule.isAppliedToProduct
            DFArule.isAppliedToProductPart = rule.isAppliedToProductPart
            DFArule.isAppliedToAssemblySequence = rule.isAppliedToAssemblySequence
            # DFArule.hasScorePropagation  = rule[6]    #TODO: check how to add
            # DFArule.hasScoreType  = rule[7]           #TODO: check how to add
            # update the preprocess data
            if DFArule.hasPreprocessedData is not None:
                DFArule.hasPreprocessesData.items[0].hasDataType = rule[8][0]
                DFArule.hasPreprocessesData.items[0].hasDataReference = rule[8][1]
            # update the options
            # TODO: Add the options - no example available to validate

def addDFArule(ruleContainer, newRule,API):
    """
      For internal use only!
    """
    # generate a new ASG config and configure
    rule = API.create_noPlatformRoot("DFA_Rule")
    rule.hasName = newRule.Name
    rule.hasDescription = newRule.Description
    rule.isAppliedToProduct = newRule.isAppliedToProduct
    rule.isAppliedToProductPart = newRule.isAppliedToProductPart
    rule.isAppliedToAssemblySequence = newRule.isAppliedToAssemblySequence
    # DFArule.hasScorePropagation  = rule[6]    #TODO:Chech how to add
    # DFArule.hasScoreType  = rule[7]           #TODO:Check how to add
    # update the preprocess data
    if rule.hasPreprocessedData is not None:
        rule.hasPreprocessesData.items[0].hasDataType = newRule[8][0]
        rule.hasPreprocessesData.items[0].hasDataReference = newRule[8][1]
        # update the options
        # TODO: Add options - no example available to validate

    try:
        ruleContainer.hasDFARule.append(rule)
    except Exception as e:
        print(e)



#interface objects
from .InterfaceObjects import Contract,Relation,Ontology,DomainVariable
#import helper functions
from .HelperFunctions import *

#-----------------------------------------------------------------------------------------------------------
#
#   Contract-Ontology extension
#
#       TODO:
#
# -----------------------------------------------------------------------------------------------------------


def getContract( ContractName,model):
    """
      Function to fetch a contract model using the name as identifier.
      The function returns an interface block of the contract model

      :param string ContractName: Name identifier of the Contract
      :param object model: Metamodel instance model

      :return object ContractModel: Interface object of the contract model (-1 = error)
    """

    contract=None
    try:
        for c in model.includesCodesignProblem.hasContracts.items:
            if c.hasName == ContractName:
                contract = Contract(Name=c.hasName,Description=c.hasDescription)
                #assumptions
                for asmpt in c.hasAssumption.items:
                    temp = []
                    for term in asmpt.expression.hasTerm.items:
                        if 'LEFT' in term.hasInfixPosition.name:
                            x = str(type(term))
                            if 'VariableTerm' in x:
                                v = DomainVariable(Name=term.hasDomainVariable.hasName,Description=term.hasDomainVariable.hasDescription,Unit=term.hasDomainVariable.hasUnit)
                                temp.append(v)
                            else:
                                temp.append(term.hasValue)
                        if 'RIGHT' in term.hasInfixPosition.name:
                            temp.append(asmpt.expression.hasExpressionOperator.hasValue)
                            x = str(type(term))
                            if 'VariableTerm' in x:
                                v = DomainVariable(Name=term.hasDomainVariable.hasName,
                                                   Description=term.hasDomainVariable.hasDescription,
                                                   Unit=term.hasDomainVariable.hasUnit)
                                temp.append(v)
                            else:
                                temp.append(term.hasValue)
                    contract.Assumptions.append(temp)


                #guarantees
                for grnt in c.hasGuarantee.items:
                    temp = []
                    for term in grnt.hasExpression.hasTerm.items:
                        if 'LEFT' in term.hasInfixPosition.name:
                            x = str(type(term))
                            if 'VariableTerm' in x:
                                v = DomainVariable(Name=term.hasDomainVariable.hasName,
                                                   Description=term.hasDomainVariable.hasDescription,
                                                   Unit=term.hasDomainVariable.hasUnit)
                                temp.append(v)
                            else:
                                temp.append(term.hasValue)
                        if 'RIGHT' in term.hasInfixPosition.name:
                            temp.append(grnt.hasExpression.hasExpressionOperator.hasValue)
                            x = str(type(term))
                            if 'VariableTerm' in x:
                                v = DomainVariable(Name=term.hasDomainVariable.hasName,
                                                   Description=term.hasDomainVariable.hasDescription,
                                                   Unit=term.hasDomainVariable.hasUnit)
                                temp.append(v)
                            else:
                                temp.append(term.hasValue)
                    contract.Guarantees.append(temp)

        return contract
    except:
        return -1



def updateContract(interfaceObject,model,KBPath):
    """
          Function to update a contract matching the name as identifier.
          The contract interface object is used to interface with the function.
          The function returns whether or not the function is performed correctly

          :param object ContractModel: Interface object of the Contract model
          :param object model: Metamodel instance model
          :param string KBPath: Absolute path to the metamodel instance model

          :return int Error: -1 = error, 1= function performed correcly
        """

    error = -1
    return error


def setContract(interfaceObject,model,KBPath):
    """
             Function to add a contract to an blank KB.
             The contract interface object is used to interface with the function.
             The function returns whether or not the function is performed correctly

             :param object ContractModel: Interface object of the Contract model
             :param object model: Metamodel instance model
             :param string KBPath: Absolute path to the metamodel instance model

             :return int Error: -1 = error, 1= function performed correcly
    """
    error = -1
    return error


def getOntology(model):
    """
      Function to fetch the complete ontology model.
      The function returns an interface block of the ontology model

      :param object model: Metamodel instance model

      :return object OntologyModel: Interface object of the ontology model (-1 = error)
    """
    ontology = None
    try:
        ontology_import = model.includesCodesignProblem.hasOntology
        ontology = Ontology(Name="Not in MM",Description="Not in MM")
        for relation in ontology_import.hasRelation.items:
            source = DomainVariable(Name=relation.hasSource.hasDomainvariable.hasName,Description=relation.hasSource.hasDomainvariable.hasDescription,Unit=relation.hasSource.hasDomainvariable.hasUnit)
            destination = DomainVariable(Name=relation.hasDestination.hasDomainvariable.hasName,Description=relation.hasDestination.hasDomainvariable.hasDescription, Unit=relation.hasDestination.hasDomainvariable.hasUnit)
            r = Relation(Name=relation.hasName, Description=relation.hasDescription, Source=source,Destination=destination)

            x = str(type(relation))
            if 'L1_FuzzyCognitiveRelation' in x:
                r.SensitivityDirection = relation.SensitivityDirection.name
            if 'L2_FuzzyCognitiveRelation' in x:
                r.SensitivityDirection = relation.SensitivityDirection.name
                r.Weight = relation.WeightIndex
            if 'L2_Relation' in x:
                r.Weight = relation.WeightIndex

            ontology.Relations.append(r)
        return ontology
    except:
        return -1





def updateOntology(interfaceObject,model,KBPath):
    """
          Function to update the complete ontology model.
          The ontology interface object is used to interface with the function.
          The function returns whether or not the function is performed correctly

          :param object OntologyModel: Interface object of the ontology model
          :param object model: Metamodel instance model
          :param string KBPath: Absolute path to the metamodel instance model

          :return int Error: -1 = error, 1= function performed correcly
        """

    error = -1
    return error


def setOntology(interfaceObject,model,KBPath):
    """
             Function to add the ontology model to a blank KB.
             The ontology interface object is used to interface with the function.
             The function returns whether or not the function is performed correctly

             :param object OntologyModel: Interface object of the ontology model
             :param object model: Metamodel instance model
             :param string KBPath: Absolute path to the metamodel instance model

             :return int Error: -1 = error, 1= function performed correcly
    """
    error = -1
    return error


from setuptools import setup

setup(
name='Interface_KB',
version='0.5.0',
author='Bert Van Acker',
author_email='bert.vanacker@uantwerpen.be',
packages=['Interface_KB'],
url='https://cosysgit.uantwerpen.be/bvanacker/knowledgebase-interface.git',
license='LICENSE.txt',
description='Knowledge-Base (KB) interface (API)',
scripts=['Interface_KB/KB_Interface.py','Interface_KB/ASGFunctions.py','Interface_KB/ContractFunctions.py','Interface_KB/DFAFunctions.py','Interface_KB/HelperFunctions.py','Interface_KB/InterfaceObjects.py','Interface_KB/PerformanceFunctions.py'],
long_description=open('README.md').read(),
install_requires=[
   "pyecore"
],
)
